#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
import argparse
import ast
import contextlib
import os
import random
import termios
import time
import timeit
import tty

import gymnasium as gym

import nle  # noqa: F401
from nle import nethack

_ACTIONS = tuple(
    [nethack.MiscAction.MORE]
    + list(nethack.CompassDirection)
    + list(nethack.CompassDirectionLonger)
)


@contextlib.contextmanager
def dummy_context():
    yield None


@contextlib.contextmanager
def no_echo():
    tt = termios.tcgetattr(0)
    try:
        tty.setraw(0)
        yield
    finally:
        termios.tcsetattr(0, termios.TCSAFLUSH, tt)


def go_back(num_lines):
    print("\033[%dA" % num_lines)


def get_action(env, is_raw_env):
    if FLAGS.mode == "random":
        if not is_raw_env:
            action = env.action_space.sample()
        else:
            action = random.choice(_ACTIONS)
            print(action)
    elif FLAGS.mode == "human":
        while True:
            with no_echo():
                ch = ord(os.read(0, 1))
            if ch in [nethack.C("c")]:
                print("Received exit code {}. Aborting.".format(ch))
                return None
            try:
                if is_raw_env:
                    action = ch
                else:
                    action = env.unwrapped.actions.index(ch)
                break
            except ValueError:
                print(
                    ("Selected action '%s' is not in action list. Please try again.")
                    % chr(ch)
                )
                if not FLAGS.print_frames_separately:
                    print("\033[2A")  # Go up 2 lines.
                continue
    return action


def play():
    is_raw_env = FLAGS.env == "raw"

    if is_raw_env:
        if FLAGS.savedir is not None:
            os.makedirs(FLAGS.savedir, exist_ok=True)
            ttyrec = os.path.join(FLAGS.savedir, "nle.ttyrec.bz2")
        else:
            ttyrec = "/dev/null"
        env = nethack.Nethack(ttyrec=ttyrec, wizard=FLAGS.wizard)
    else:
        env = gym.make(
            FLAGS.env,
            save_ttyrec_every=2,
            savedir=FLAGS.savedir,
            max_episode_steps=FLAGS.max_steps,
            allow_all_yn_questions=True,
            allow_all_modes=True,
            wizard=FLAGS.wizard,
            render_mode=FLAGS.render_mode,
        )
        if FLAGS.seeds is not None:
            env.unwrapped.seed(FLAGS.seeds)

    obs, reset_info = env.reset()

    steps = 0
    episodes = 0
    reward = 0.0
    action = None

    mean_sps = 0
    mean_reward = 0.0

    total_start_time = timeit.default_timer()
    start_time = total_start_time

    while True:
        if not FLAGS.no_render:
            if not is_raw_env:
                print("-" * 8 + " " * 71)
                print(f"Previous reward: {str(reward):64s}")
                act_str = (
                    repr(env.unwrapped.actions[action]) if action is not None else ""
                )
                print(f"Previous action: {str(act_str):64s}")
                print("-" * 8)
                env.render()
                print("-" * 8)
                print(obs["blstats"])
                if not FLAGS.print_frames_separately:
                    go_back(num_lines=33)
            else:
                print("Previous action:", action)
                obs = dict(zip(nle.nethack.OBSERVATION_DESC.keys(), obs))
                print(
                    nle.nethack.tty_render(
                        obs["tty_chars"], obs["tty_colors"], obs["tty_cursor"]
                    )
                )
                if not FLAGS.print_frames_separately:
                    go_back(num_lines=len(obs["tty_chars"]) + 3)

        action = get_action(env, is_raw_env)

        if action is None:
            break

        if is_raw_env:
            obs, done = env.step(action)
        else:
            obs, reward, done, truncated, info = env.step(action)
        steps += 1

        if is_raw_env:
            done = done or steps >= FLAGS.max_steps  # NLE does this by default.
        else:
            mean_reward += (reward - mean_reward) / steps

        if not done:
            continue

        time_delta = timeit.default_timer() - start_time

        if not is_raw_env:
            print("Final reward:", reward)
            print("End status:", info["end_status"].name)
            print("Mean reward:", mean_reward)

        sps = steps / time_delta
        print("Episode: %i. Steps: %i. SPS: %f" % (episodes, steps, sps))

        episodes += 1
        mean_sps += (sps - mean_sps) / episodes

        start_time = timeit.default_timer()

        steps = 0
        mean_reward = 0.0

        if episodes == FLAGS.ngames:
            break
        env.reset()
    env.close()
    print(
        "Finished after %i episodes and %f seconds. Mean sps: %f"
        % (episodes, timeit.default_timer() - total_start_time, mean_sps)
    )


def main():
    parser = argparse.ArgumentParser(description="NLE Play tool.")
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Enables debug mode, which will drop stack into "
        "an ipdb shell if an exception is raised.",
    )
    parser.add_argument(
        "-m",
        "--mode",
        type=str,
        default="human",
        choices=["human", "random"],
        help="Control mode. Defaults to 'human'.",
    )
    parser.add_argument(
        "-e",
        "--env",
        type=str,
        default="NetHackScore-v0",
        help="Gym environment spec. Defaults to 'NetHackStaircase-v0'.",
    )
    parser.add_argument(
        "-n",
        "--ngames",
        type=int,
        default=1,
        help="Number of games to be played before exiting. "
        "NetHack will auto-restart if > 1.",
    )
    parser.add_argument(
        "--max-steps",
        type=int,
        default=1_000_000,
        help="Number of maximum steps per episode.",
    )
    parser.add_argument(
        "--seeds",
        default=None,
        help="Seeds to send to NetHack. Can be a dict or int. "
        "Defaults to None (no seeding).",
    )
    parser.add_argument(
        "--savedir",
        default="nle_data/play_data",
        help="Directory path where data will be saved. "
        "Defaults to 'nle_data/play_data'.",
    )
    parser.add_argument(
        "--no-render", action="store_true", help="Disables env.render()."
    )
    parser.add_argument(
        "--render_mode",
        type=str,
        default="human",
        choices=["human", "full", "ansi"],
        help="Render mode. Defaults to 'human'.",
    )
    parser.add_argument(
        "--print-frames-separately",
        "-p",
        action="store_true",
        help="Don't overwrite frames, print them all.",
    )
    parser.add_argument(
        "--wizard",
        "-D",
        action="store_true",
        help="Use wizard mode.",
    )
    global FLAGS
    FLAGS = parser.parse_args()

    if FLAGS.debug:
        import ipdb

        cm = ipdb.launch_ipdb_on_exception
    else:
        cm = dummy_context

    with cm():
        if FLAGS.seeds is not None:
            # to handle both int and dicts
            FLAGS.seeds = ast.literal_eval(FLAGS.seeds)

        if FLAGS.savedir == "args":
            FLAGS.savedir = "{}_{}_{}.zip".format(
                time.strftime("%Y%m%d-%H%M%S"), FLAGS.mode, FLAGS.env
            )
        elif FLAGS.savedir == "None":
            FLAGS.savedir = None  # Not saving any ttyrecs.

        play()


if __name__ == "__main__":
    main()



# Copyright (c) Facebook, Inc. and its affiliates.

import collections
import pprint
import sys

Entry = collections.namedtuple("Entry", "size hash line file")


def main():
    allocs = {}
    with open(sys.argv[1]) as heaplog:
        for line in heaplog:
            entries = line.split()
            if entries[0] == "+":
                entry = Entry(*entries[1:])
                allocs[entry.hash] = entry
            else:
                entry = Entry(None, *entries[1:])
                if entry.hash not in allocs:
                    print("dealloc not found in allocs:", line)
                    continue
                del allocs[entry.hash]

    pprint.pprint(allocs)


if __name__ == "__main__":
    main()



import contextlib
import logging
import os
import sqlite3
import time

DB = "ttyrecs.db"
logger = logging.getLogger("db")


@contextlib.contextmanager
def db(conn=None, filename=DB, new=False, rw=None, **kwargs):
    if conn:
        yield conn
        return
    try:
        conn = connect(filename, new, rw, **kwargs)
        yield conn
    finally:
        if conn:
            conn.close()


def exists(filename=DB):
    return os.path.exists(filename)


def connect(filename=DB, new=False, rw=None, **kwargs):
    assert new ^ os.path.exists(filename)
    if not new and not rw:
        filename += "?mode=ro"
    return sqlite3.connect("file:" + filename, uri=True, **kwargs)


def ls(conn=None):
    with db(conn) as conn:
        c = conn.cursor()
        c.execute("SELECT * FROM meta")
        ctime, mtime = c.fetchone()
        logger.info(
            time.ctime(ctime),
            time.ctime(mtime),
        )
        for row in c.execute("SELECT rowid, path FROM ttyrecs ORDER BY rowid"):
            yield row


def vacuum(conn=None):
    with db(conn=conn, rw=True) as conn:
        conn.execute("VACUUM")


def get_row(rowid="1", conn=None):
    with db(conn) as conn:
        result = conn.execute(
            "SELECT rowid, * FROM ttyrecs WHERE rowid = ?", (rowid,)
        ).fetchone()
        if result is None:
            raise ValueError("Row %s not found" % rowid)
        return result


def get_meta(conn=None):
    with db(conn) as conn:
        return conn.execute("SELECT * FROM meta").fetchone()


def get_root(dataset_name, conn=None):
    with db(conn) as conn:
        return conn.execute(
            "SELECT root FROM roots WHERE dataset_name=?", (dataset_name,)
        ).fetchone()[0]


def set_root(dataset_name, root):
    root = os.path.abspath(root)
    with db(rw=True) as conn:
        conn.execute(
            "UPDATE roots SET root = ? WHERE dataset_name = ?", (root, dataset_name)
        )
        conn.execute("UPDATE meta SET mtime = ?", (time.time(),))
        conn.commit()


def get_ttyrec_version(dataset_name, conn=None):
    with db(conn) as conn:
        return conn.execute(
            "SELECT ttyrec_version FROM roots WHERE dataset_name=?", (dataset_name,)
        ).fetchone()[0]


def get_most_recent_games(n=1, conn=None):
    with db(conn=conn) as conn:
        c = conn.execute("SELECT gameid FROM games ORDER BY gameid DESC LIMIT ?", (n,))
        return [g[0] for g in c.fetchall()]


def count_games(dataset_name, conn=None):
    with db(conn=conn) as conn:
        c = conn.execute(
            "SELECT COUNT(datasets.gameid) FROM datasets WHERE datasets.dataset_name=?",
            (dataset_name,),
        )
        return c.fetchone()[0]


def get_games(dataset_name, conn=None):
    with db(conn=conn) as conn:
        c = conn.execute(
            "SELECT games.name, games.gameid, games.starttime, games.endtime "
            "FROM games "
            "JOIN datasets ON games.gameid=datasets.gameid "
            "WHERE datasets.dataset_name=?",
            (dataset_name,),
        )
        for row in c:
            yield row


def add_games(dataset_name, *gameids, conn=None, commit=True):
    with db(conn, rw=True) as conn:
        conn.executemany(
            "INSERT INTO datasets VALUES (?, ?)",
            zip(gameids, [dataset_name] * len(gameids)),
        )
        conn.execute("UPDATE meta SET mtime = ?", (time.time(),))
        if commit:
            conn.commit()


def drop_games(dataset_name, *gameids, conn=None, commit=True):
    with db(conn, rw=True) as conn:
        conn.executemany(
            "DELETE FROM datasets WHERE gameid=? AND dataset_name=?",
            zip(gameids, [dataset_name] * len(gameids)),
        )
        conn.execute("UPDATE meta SET mtime = ?", (time.time(),))
        if commit:
            conn.commit()


def delete_games_with_select(select, not_in=False, conn=None, commit=True):
    _not = "NOT" if not_in else ""
    with db(conn, rw=True) as conn:
        conn.execute("DELETE FROM ttyrecs WHERE %s gameid IN (%s)" % (_not, select))
        conn.execute("DELETE FROM datasets WHERE %s gameid IN (%s)" % (_not, select))
        conn.execute("DELETE FROM games WHERE %s gameid IN (%s)" % (_not, select))
        conn.execute("UPDATE meta SET mtime = ?", (time.time(),))
        if commit:
            conn.commit()


def create_dataset(dataset_name, root, ttyrec_version=0, conn=None, commit=True):
    with db(conn, rw=True) as conn:
        conn.execute(
            "INSERT INTO roots VALUES (?, ?, ?)", (dataset_name, root, ttyrec_version)
        )
        conn.execute("UPDATE meta SET mtime = ?", (time.time(),))
        if commit:
            conn.commit()


def delete_dataset(dataset_name, conn=None, commit=True):
    with db(conn, rw=True) as conn:
        conn.execute("DELETE datasets WHERE dataset_name=?", (dataset_name,))
        conn.execute("DELETE roots WHERE dataset_name=?", (dataset_name,))
        conn.execute("UPDATE meta SET mtime = ?", (time.time(),))
        if commit:
            conn.commit()


def create(filename=DB):
    ctime = time.time()

    with db(filename=filename, new=True) as conn:
        logger.info("Creating '%s' ...", DB)
        c = conn.cursor()

        c.execute(
            """CREATE TABLE meta
            (
                ctime REAL,
                mtime REAL
            )"""
        )
        c.execute("INSERT INTO meta VALUES (?, ?)", (ctime, ctime))

        c.execute(
            """CREATE TABLE ttyrecs
            (
                path        TEXT,
                part        INTEGER,
                size        INTEGER,
                mtime       REAL,
                gameid      INTEGER,
                PRIMARY KEY (gameid, part, path)
            )"""
        )

        c.execute(
            """CREATE TABLE games
            (
                gameid      INTEGER PRIMARY KEY,
                version     TEXT,
                points      INTEGER,
                deathdnum   INTEGER,
                deathlev    INTEGER,
                maxlvl      INTEGER,
                hp          INTEGER,
                maxhp       INTEGER,
                deaths      INTEGER,
                deathdate   INTEGER,
                birthdate   INTEGER,
                uid         INTEGER,
                role        TEXT,
                race        TEXT,
                gender      TEXT,
                align       TEXT,
                name        TEXT,
                death       TEXT,
                conduct     TEXT,
                turns       INTEGER,
                achieve     TEXT,
                realtime    INTEGER,
                starttime   INTEGER,
                endtime     INTEGER,
                gender0     TEXT,
                align0      TEXT,
                flags       TEXT
            )
            """
        )

        c.execute(
            """CREATE TABLE datasets
            (
                gameid        INTEGER,
                dataset_name  TEXT,
                PRIMARY KEY   (dataset_name, gameid)
            )
            """
        )

        c.execute(
            """CREATE TABLE roots
            (
                dataset_name   TEXT PRIMARY KEY,
                root           TEXT,
                ttyrec_version INTEGER
            )"""
        )

        conn.commit()
    logger.info(
        "Created Empty '%s'. Size: %.2f MB",
        filename,
        os.path.getsize(filename) / 1024**2,
    )



# Copyright (c) Facebook, Inc. and its affiliates.
import multiprocessing as mp
import queue
import random
import threading

import gymnasium as gym
import pytest

import nle  # noqa: F401

START_METHODS = [m for m in ("fork", "spawn") if m in mp.get_all_start_methods()]


def new_env_one_step():
    env = gym.make("NetHackScore-v0")
    env.reset()
    obs, reward, terminated, _, _ = env.step(0)
    return terminated


@pytest.mark.parametrize(
    "ctx", [mp.get_context(m) for m in START_METHODS], ids=START_METHODS
)
class TestEnvSubprocess:
    def test_env_in_subprocess(self, ctx):
        p = ctx.Process(target=new_env_one_step)
        p.start()
        p.join()
        assert p.exitcode == 0

    def test_env_before_and_in_subprocess(self, ctx):
        new_env_one_step()
        p = ctx.Process(target=new_env_one_step)
        p.start()
        p.join()
        assert p.exitcode == 0


ACTIONS = [0, 1, 2]


class TestParallelEnvs:
    def test_two_nles(self):
        envs = [gym.make("NetHackScore-v0") for _ in range(2)]

        env, *queue = envs
        env.reset()

        num_resets = 1

        while num_resets < 4:
            _, _, terminated, _, _ = env.step(random.choice(ACTIONS))
            if terminated:
                queue.append(env)
                env = queue.pop(0)
                env.reset()
                num_resets += 1

    def test_threaded_nles(self, num_envs=10, num_threads=3):
        readyqueue = queue.Queue()
        resetqueue = queue.Queue()

        def target():
            while True:
                env = resetqueue.get()
                if env is None:
                    return
                env.reset()
                readyqueue.put(env)

        threads = [threading.Thread(target=target) for _ in range(num_threads)]
        for t in threads:
            t.start()

        envs = [gym.make("NetHackScore-v0") for _ in range(num_envs)]
        for env in envs:
            resetqueue.put(env)
        env = readyqueue.get()

        num_resets = 1

        while num_resets < 4:
            a = random.choice(ACTIONS)
            _, _, terminated, _, _ = env.step(a)
            if terminated:
                resetqueue.put(env)
                env = readyqueue.get()
                num_resets += 1

        for _ in threads:
            resetqueue.put(None)

        for t in threads:
            t.join()



# Copyright (c) Facebook, Inc. and its affiliates.
from nle.nethack.actions import *  # noqa: F403
from nle._pynethack.nethack import *  # noqa: F403
from nle.nethack.nethack import (
    Nethack,
    NETHACKOPTIONS,
    DUNGEON_SHAPE,
    BLSTATS_SHAPE,
    MESSAGE_SHAPE,
    INV_SIZE,
    PROGRAM_STATE_SHAPE,
    INTERNAL_SHAPE,
    OBSERVATION_DESC,
    TTYREC_VERSION,
    tty_render,
)



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
import os
import pathlib
import random
import sys
import tempfile

import gymnasium as gym
import numpy as np
import pytest

import nle
import nle.env
from nle import nethack


def get_nethack_env_ids():
    specs = gym.envs.registry.keys()
    # Ignoring base environment, since we can't handle random actions yet with
    # the full action space, and this requires a whole different set of tests.
    # For now this is OK, since NetHackScore-v0 is very similar.
    return [
        spec for spec in specs if spec.startswith("NetHack") and spec != "NetHack-v0"
    ]


def rollout_env(env, max_rollout_len):
    """Produces a rollout and asserts step outputs.

    Returns final reward. Does not assume that the environment has already been
    reset.
    """
    obs, reset_info = env.reset()
    assert env.observation_space.contains(obs)

    for _ in range(max_rollout_len):
        a = env.action_space.sample()
        obs, reward, terminated, truncated, info = env.step(a)
        assert env.observation_space.contains(obs)
        assert isinstance(reward, float)
        assert isinstance(terminated, bool)
        assert isinstance(info, dict)
        if terminated:
            assert not info["is_ascended"]
            break
    env.close()
    return reward


def term_screen(obs):
    return "\n".join("".join(chr(c) for c in row) for row in obs["tty_chars"])


def compare_rollouts(env0, env1, max_rollout_len):
    """Checks that two active environments return the same rollout.

    Assumes that the environments have already been reset.
    """
    step = 0
    while True:
        a = env0.action_space.sample()
        obs0, reward0, terminated0, truncated0, info0 = env0.step(a)
        obs1, reward1, terminated1, truncated1, info1 = env1.step(a)
        step += 1

        s0, s1 = term_screen(obs0), term_screen(obs1)
        top_ten_msg = "You made the top ten list!"
        if top_ten_msg in s0:
            assert top_ten_msg in s1
        else:
            np.testing.assert_equal(obs0, obs1)
        assert reward0 == reward1
        assert terminated0 == terminated1
        assert truncated0 == truncated1
        assert info0 == info1

        if terminated0 or step >= max_rollout_len:
            return


@pytest.mark.parametrize("env_name", get_nethack_env_ids())
@pytest.mark.parametrize("wizard", [False, True])
class TestGymEnv:
    @pytest.fixture(autouse=True)  # will be applied to all tests in class
    def make_cwd_tmp(self, tmpdir):
        """Makes cwd point to the test's tmpdir."""
        with tmpdir.as_cwd():
            yield

    def test_init(self, env_name, wizard):
        """Tests default initialization given standard env specs."""
        env = gym.make(env_name, wizard=wizard)
        del env

    def test_reset(self, env_name, wizard):
        """Tests default initialization given standard env specs."""
        env = gym.make(env_name, wizard=wizard)
        obs, reset_info = env.reset()
        assert env.observation_space.contains(obs)

    def test_chars_colors_specials(self, env_name, wizard):
        env = gym.make(
            env_name, observation_keys=("chars", "colors", "specials", "blstats")
        )
        obs, reset_info = env.reset()

        assert "specials" in obs
        x, y = obs["blstats"][:2]

        # That's where you're @.
        assert obs["chars"][y, x] == ord("@")

        # You're bright (4th bit, 8) white (7), too.
        assert obs["colors"][y, x] == 8 ^ 7

    def test_default_wizard_mode(self, env_name, wizard):
        if wizard:
            if env_name.startswith("NetHackChallenge-"):
                pytest.skip("No wizard mode in NetHackChallenge")
            env = gym.make(env_name, wizard=wizard)
            assert "playmode:debug" in env.unwrapped.nethack.options
        else:
            # do not send a parameter to test a default
            env = gym.make(env_name)
            assert "playmode:debug" not in env.unwrapped.nethack.options


class TestWizardMode:
    def test_wizlevelport(self):
        actions = (
            list(nethack.USEFUL_ACTIONS)
            + list(nethack.TextCharacters)
            + list(nethack.WizardCommand)
        )

        env = gym.make(
            "NetHack-v0",
            wizard=True,
            actions=actions,
            allow_all_yn_questions=True,
            allow_all_modes=True,
        )
        env.reset()
        env.step(actions.index(nethack.WizardCommand.WIZLEVELPORT))
        for c in b"10\r":
            env.step(actions.index(c))


class TestWizkit:
    @pytest.fixture(autouse=True)  # will be applied to all tests in class
    def make_cwd_tmp(self, tmpdir):
        """Makes cwd point to the test's tmpdir."""
        with tmpdir.as_cwd():
            yield

    def test_meatball_exists(self):
        """Test loading stuff via wizkit"""
        env = gym.make("NetHack-v0", wizard=True)
        found = dict(meatball=0)
        obs, reset_info = env.reset(options={"wizkit_items": list(found.keys())})
        for line in obs["inv_strs"]:
            if np.all(line == 0):
                break
            for key in found:
                if key in line.tobytes().decode("utf-8"):
                    found[key] += 1
        for key, count in found.items():
            assert key == key and count > 0
        del env

    def test_wizkit_no_wizard_mode(self):
        env = gym.make("NetHack-v0", wizard=False)
        with pytest.raises(ValueError) as e_info:
            env.reset(options={"wizkit_items": ["meatball"]})
        assert e_info.value.args[0] == "Set wizard=True to use the wizkit option."

    def test_wizkit_file(self):
        env = gym.make("NetHack-v0", wizard=True)
        req_items = ["meatball", "apple"]
        env.reset(options={"wizkit_items": req_items})

        # TODO: Test inventory here.
        env.reset(options={"wizkit_items": req_items})
        del env


@pytest.mark.parametrize("env_name", [e for e in get_nethack_env_ids() if "Score" in e])
class TestBasicGymEnv:
    def test_inventory(self, env_name):
        env = gym.make(
            env_name,
            observation_keys=(
                "chars",
                "inv_glyphs",
                "inv_strs",
                "inv_letters",
                "inv_oclasses",
            ),
        )
        obs, reset_info = env.reset()

        found = dict(spellbook=0, apple=0)
        for line in obs["inv_strs"]:
            if np.all(line == 0):
                break
            for key in found:
                if key in line.tobytes().decode("utf-8"):
                    found[key] += 1

        for key, count in found.items():
            assert key == key and count > 0

        assert "inv_strs" in obs

        index = 0
        if obs["inv_letters"][index] != ord("a"):
            # We autopickedup some gold.
            assert obs["inv_letters"][index] == ord("$")
            assert obs["inv_oclasses"][index] == nethack.COIN_CLASS
            index = 1

        assert obs["inv_letters"][index] == ord("a")
        assert obs["inv_oclasses"][index] == nethack.ARMOR_CLASS


@pytest.mark.parametrize("env_name", get_nethack_env_ids())
@pytest.mark.parametrize("rollout_len", [500])
class TestGymEnvRollout:
    @pytest.fixture(autouse=True)  # will be applied to all tests in class
    def make_cwd_tmp(self, tmpdir):
        """Makes cwd point to the test's tmpdir."""
        with tmpdir.as_cwd():
            yield

    def test_rollout(self, env_name, rollout_len):
        """Tests rollout_len steps (or until termination) of random policy."""
        with tempfile.TemporaryDirectory() as savedir:
            env = gym.make(env_name, save_ttyrec_every=1, savedir=savedir)
            rollout_env(env, rollout_len)
            env.close()

            assert os.path.exists(
                os.path.join(
                    savedir,
                    "nle.%i.0.ttyrec%i.bz2" % (os.getpid(), nethack.TTYREC_VERSION),
                )
            )
            assert os.path.exists(
                os.path.join(savedir, "nle.%i.xlogfile" % os.getpid())
            )

    def test_rollout_no_archive(self, env_name, rollout_len):
        """Tests rollout_len steps (or until termination) of random policy."""
        env = gym.make(env_name, savedir=None)
        assert env.unwrapped.savedir is None
        rollout_env(env, rollout_len)

    def test_seed_interface_output(self, env_name, rollout_len):
        """Tests whether env.seed output can be reused correctly."""
        if not nethack.NLE_ALLOW_SEEDING:
            return  # Nothing to test.
        if env_name.startswith("NetHackChallenge"):
            pytest.skip("Not running seed test on NetHackChallenge")

        env0 = gym.make(env_name)
        env1 = gym.make(env_name)

        seed_list0 = env0.unwrapped.seed()
        env0.reset()

        assert env0.unwrapped.get_seeds() == seed_list0

        seed_list1 = env1.unwrapped.seed(*seed_list0)
        assert seed_list0 == seed_list1

    def test_seed_rollout_seeded(self, env_name, rollout_len):
        """Tests that two seeded envs return same step data."""
        if not nethack.NLE_ALLOW_SEEDING:
            return  # Nothing to test.
        if env_name.startswith("NetHackChallenge"):
            pytest.skip("Not running seed test on NetHackChallenge")

        env0 = gym.make(env_name)
        env1 = gym.make(env_name)

        env0.unwrapped.seed(123456, 789012)
        obs0 = env0.reset()
        seeds0 = env0.unwrapped.get_seeds()

        assert seeds0 == (123456, 789012, False)

        env1.unwrapped.seed(*seeds0)
        obs1 = env1.reset()
        seeds1 = env1.unwrapped.get_seeds()

        assert seeds0 == seeds1

        np.testing.assert_equal(obs0, obs1)
        compare_rollouts(env0, env1, rollout_len)

    def test_seed_rollout_seeded_int(self, env_name, rollout_len):
        """Tests that two seeded envs return same step data."""
        if not nethack.NLE_ALLOW_SEEDING:
            return  # Nothing to test.
        if env_name.startswith("NetHackChallenge"):
            pytest.skip("Not running seed test on NetHackChallenge")

        env0 = gym.make(env_name)
        env1 = gym.make(env_name)

        initial_seeds = (
            random.randrange(sys.maxsize),
            random.randrange(sys.maxsize),
            False,
        )
        env0.unwrapped.seed(*initial_seeds)
        obs0 = env0.reset()
        seeds0 = env0.unwrapped.get_seeds()

        env1.unwrapped.seed(*seeds0)
        obs1 = env1.reset()
        seeds1 = env1.unwrapped.get_seeds()

        assert seeds0 == seeds1 == initial_seeds

        np.testing.assert_equal(obs0, obs1)
        compare_rollouts(env0, env1, rollout_len)

    def test_render_ansi(self, env_name, rollout_len):
        env = gym.make(env_name, render_mode="ansi")
        env.reset()
        for _ in range(rollout_len):
            action = env.action_space.sample()
            _, _, terminated, _, _ = env.step(action)
            if terminated:
                env.reset()
            output = env.render()
            assert isinstance(output, str)
            assert len(output.replace("\n", "")) == np.prod(nle.env.DUNGEON_SHAPE)


class TestGymDynamics:
    """Tests a few game dynamics."""

    @pytest.fixture(autouse=True)  # Will be applied to all tests in class.
    def make_cwd_tmp(self, tmpdir):
        """Makes cwd point to the test's tmpdir."""
        with tmpdir.as_cwd():
            yield

    @pytest.fixture
    def env(self):
        e = gym.make("NetHackScore-v0")
        try:
            yield e
        finally:
            e.close()

    def test_kick_and_quit(self, env):
        env.reset()
        kick = env.unwrapped.actions.index(nethack.Command.KICK)
        obs, reward, terminated, _, _ = env.step(kick)
        assert b"In what direction? " in bytes(obs["message"])
        env.step(nethack.MiscAction.MORE)

        # Hack to quit.
        env.unwrapped.nethack.step(nethack.M("q"))
        obs, reward, terminated, _, _ = env.step(env.unwrapped.actions.index(ord("y")))

        assert terminated
        assert reward == 0.0

    def test_final_reward(self, env):
        obs, reset_info = env.reset()

        for _ in range(100):
            obs, reward, terminated, _, info = env.step(env.action_space.sample())
            if terminated:
                break

        if terminated:
            assert reward == 0.0
            return

        # Hopefully, we got some positive reward by now.

        # Get out of any menu / yn_function.
        env.step(env.unwrapped.actions.index(ord("\r")))

        # Hack to quit.
        env.unwrapped.nethack.step(nethack.M("q"))
        _, reward, terminated, _, info = env.step(env.unwrapped.actions.index(ord("y")))

        assert terminated
        assert reward == 0.0

    def test_ttyrec_every(self):
        path = pathlib.Path(".")
        env = gym.make("NetHackChallenge-v0", save_ttyrec_every=2, savedir=str(path))
        pid = os.getpid()
        for episode in range(10):
            env.reset()
            for c in [ord(" "), ord(" "), ord("<"), ord("y")]:
                _, _, terminated, *_ = env.step(env.unwrapped.actions.index(c))
            assert terminated

            if episode % 2 != 0:
                continue
            contents = {str(p) for p in path.iterdir()}
            # `contents` includes xlogfile and ttyrecs.
            assert len(contents) - 1 == episode // 2 + 1
            assert (
                "nle.%i.%i.ttyrec%i.bz2" % (pid, episode, nethack.TTYREC_VERSION)
                in contents
            )
            assert "nle.%i.xlogfile" % pid in contents

        with open("nle.%i.xlogfile" % pid, "r") as f:
            entries = f.readlines()

        assert len(entries) == 10

    def test_env_truncation(self):
        test_horizon = 10

        env = gym.make("NetHack-v0", max_episode_steps=test_horizon)
        env.reset()
        for _steps in range(test_horizon - 1):
            obs, reward, termination, truncation, info = env.step(
                nethack.MiscDirection.WAIT
            )
            assert not termination
            assert not truncation

        obs, reward, termination, truncation, info = env.step(
            nethack.MiscDirection.WAIT
        )
        assert not termination
        assert truncation


class TestEnvMisc:
    """Tests miscellaneous enviroment behavior."""

    @pytest.fixture
    def env(self):
        if sys.version_info < (3, 8):
            e = gym.make("NetHackScore-v0")
        else:
            # gym 0.24+ doesnt like the shape of our observations.
            e = gym.make("NetHackScore-v0")
        try:
            yield e
        finally:
            e.close()

    def test_no_reset(self, env):
        with pytest.raises(RuntimeError, match="step called without reset()"):
            env.step(0)


class TestNetHackChallenge:
    def test_no_seed_setting(self):
        env = gym.make("NetHackChallenge-v0")
        with pytest.raises(
            RuntimeError, match="NetHackChallenge doesn't allow seed changes"
        ):
            env.unwrapped.seed()
        with pytest.raises(RuntimeError, match="Should not try changing seeds"):
            env.unwrapped.nethack.set_initial_seeds(0, 0, True)

        if not nethack.NLE_ALLOW_SEEDING:
            with pytest.raises(RuntimeError, match="Seeding not enabled"):
                env.unwrapped.nethack._pynethack.set_initial_seeds(0, 0, True)



import collections
import datetime
import glob
import os
import re
import time
from functools import partial

from nle import dataset as nld

XLOGFILE_COLUMNS = [
    ("version", str),
    ("points", int),
    ("deathdnum", int),
    ("deathlev", int),
    ("maxlvl", int),
    ("hp", int),
    ("maxhp", int),
    ("deaths", int),
    ("deathdate", int),
    ("birthdate", int),
    ("uid", int),
    ("role", str),
    ("race", str),
    ("gender", str),
    ("align", str),
    ("name", str),
    ("death", str),
    ("conduct", str),
    ("turns", int),
    ("achieve", str),
    ("realtime", int),
    ("starttime", int),
    ("endtime", int),
    ("gender0", str),
    ("align0", str),
    ("flags", str),
]

FIVE_MINS = 5 * 60
ALT_TIMEFMT = re.compile(r"(.*\.\d\d)_(\d\d)_(\d\d.*)")


def altorg_filename_to_timestamp(filename):
    ts = filename.split("/")[-1][:-11]
    # We accept time format HH_MM_SS or HH:MM:SS, but convert for ISO format.
    ts = ALT_TIMEFMT.sub(r"\1:\2:\3", ts)
    try:
        ts = datetime.datetime.fromisoformat(ts)
    except AttributeError:
        # Python 3.6 doesnt have fromisoformat
        this_date, this_time = ts.split(".")
        this_date = [int(x) for x in this_date.split("-")]
        this_time = [int(x) for x in this_time.split(":")]
        this_datetime = this_date + this_time
        ts = datetime.datetime(*this_datetime)
    except ValueError:
        print("Skipping: '%s'" % filename)
        return -1
    return ts.replace(tzinfo=datetime.timezone.utc).timestamp()


def assign_ttyrecs_to_games(ttyrecs, games):
    """Algorithm to assign a players ttyrecs to their games, knowing that only one game
    can be played at any one time."""
    assigned = []  # (path, file_creationtime, gameid, game_starttime, game_endtime)
    for t in ttyrecs:
        s_time = altorg_filename_to_timestamp(t)
        assigned.append([t, s_time, -1, -1, -1])

    # We sort games and ttyrecs by start/creationtime and will pair them in one pass.
    # NB: We only want ttyrecs that can be assigned to a game.
    assigned.sort(key=lambda x: x[1])
    games.sort(key=lambda x: x[1])

    gg, tt = 0, 0
    while gg < len(games) and tt < len(assigned):
        if assigned[tt][1] > games[gg][2]:
            # This ttyrec was created after this game ends, so check next game.
            gg += 1
        elif assigned[tt][1] < games[gg][1] - FIVE_MINS:
            # We allow a 5min window, since the ttyrec creation time and xlogfile
            # starttime can differ slightly. However ttyrec starts well before
            # this game start, so check next ttyrec.

            # Design choice: We will add the ttyrec with a -ve gameid
            # so that it will not be picked up when selecting by dataset="altorg"
            # but will still exist in the database
            assigned[tt][2] = -games[gg][0]
            assigned[tt][3] = games[gg][1]
            assigned[tt][4] = games[gg][2]
            tt += 1
        else:
            # This ttyrec starts after the game starts ( - 5 mins buffer)
            # and ends before this game ends, so we can assign
            assigned[tt][2] = games[gg][0]
            assigned[tt][3] = games[gg][1]
            assigned[tt][4] = games[gg][2]
            tt += 1
    return [(ttyrec, gameid) for ttyrec, _, gameid, _, _ in assigned if gameid != -1]


def add_altorg_directory(path, name, filename=nld.db.DB):
    """This function can be used to add the `altorg` dataset to a database.

    Once the altorg dataset has been downloaded, this function will parse its
    contents and create the dataset.

    The altorg directory structure should look like:

        altorg/
        ├── user1/
        │   ├── 2019-07-25.22:03:29.ttyrec.bz2
        │   └── 2019-07-30.16:06:23.ttyrec.bz2
        ├── user2/
        │   ├── 2019-07-25.22:03:29.ttyrec.bz2
        │   └── 2019-07-30.16:06:23.ttyrec.bz2
        ...
        ├── xlogfile.1
        ├── xlogfile.2
        ...
        ├── about.txt
        └── blacklist.txt

    Note that unlike `nle` ttyrecs, altorg episodes may be split into parts.
    We use a simple algorithm based on the file creation times and xlogfile times
    for the start and end of games to try to assign ttyrecs to games, knowing
    a player can only ever be playing on game one altorg at a time.

    This algorithm should be deterministic and always return the same dataset
    from an empty database, regardless of environment.

    """

    with nld.db.db(filename=filename, rw=True) as conn:
        print("Adding dataset '%s' ('%s') to '%s' " % (name, path, filename))

        root = os.path.abspath(path)
        stime = time.time()
        c = conn.cursor()

        # 1. Check if the dataset name exists, and add the root.
        # NB: alt.org ttyrecs are version 1, and have suffix "ttyrec.bz2"
        nld.db.create_dataset(name, root, ttyrec_version=1, conn=c, commit=False)

        # 2. Add games from xlogfile to `games` table, then `datasets` table.
        for xlogfile in sorted(
            glob.iglob(str(os.path.join(path, "xlogfile.*"))), reverse=True
        ):
            sep = ":" if xlogfile.endswith(".txt") else "\t"
            game_gen = game_data_generator(xlogfile, separator=sep)
            insert_sql = f"""
                INSERT INTO games
                VALUES (NULL, {','.join('?' for _ in XLOGFILE_COLUMNS)} )
            """
            c.executemany(insert_sql, game_gen)

            gameids = nld.db.get_most_recent_games(c.rowcount, conn=c)
            nld.db.add_games(name, *gameids, conn=c, commit=False)
            print("Found %i games in '%s'" % (len(gameids), xlogfile))

        # 3. Find all the (unblacklisted) ttyrecs belonging to each player
        #    and all the games belonging to each player.
        with open(os.path.join(path, "blacklist.txt"), "r") as f:
            blacklisted_ttyrecs = {str(os.path.join(path, p)) for p in f.readlines()}

        ttyrecs_dict = collections.defaultdict(list)
        for ttyrec in glob.iglob(path + "/*/*.ttyrec.bz2"):
            if ttyrec in blacklisted_ttyrecs:
                continue
            ttyrecs_dict[ttyrec.split("/")[-2].lower()].append(ttyrec)

        games_dict = collections.defaultdict(list)
        for pname, gameid, start, end in nld.db.get_games(name, conn=c):
            games_dict[pname.lower()].append((gameid, start, end))

        # 4. Attempt assign each player's ttyrecs to each player's games.
        #    If successful, insert into `ttyrecs` table
        print("Matching up ttyrecs to games...")
        empty_games = []
        for pname in ttyrecs_dict.keys():
            assigned = assign_ttyrecs_to_games(ttyrecs_dict[pname], games_dict[pname])
            if assigned:
                ttyrecs, gameids = zip(*assigned)
                ttyrec_gen = ttyrec_data_generator(ttyrecs, gameids, root)
                c.executemany("INSERT INTO ttyrecs VALUES (?,?,?,?,?)", ttyrec_gen)
            elif games_dict[pname]:
                empty_games.extend(gid for gid, _, _ in games_dict[pname])
        for pname in games_dict:
            if pname not in ttyrecs_dict:
                empty_games.extend(gid for gid, _, _ in games_dict[pname])

        pseudonyms = []
        for i, pname in enumerate(games_dict):
            for gameid, _, _ in games_dict[pname]:
                pseudonyms.append((f"Player{i}", gameid, pname))
        pseudnymize = "UPDATE games SET name=? WHERE gameid=? AND lower(name)=?"
        c.executemany(pseudnymize, pseudonyms)

        # 5. Purge 'empty' games from `datasets` and `games` table.
        games_with_ttyrecs = """SELECT DISTINCT(gameid) FROM ttyrecs"""
        nld.db.delete_games_with_select(
            games_with_ttyrecs, not_in=True, conn=conn, commit=False
        )

        # 6. Purge short games where the player quit almost immediately.
        start_scummed_games = """
           SELECT gameid FROM games
           WHERE (turns <=10 AND (death = "escaped" OR death ="quit")) OR turns<=0"""
        nld.db.delete_games_with_select(start_scummed_games, conn=conn, commit=False)

        mtime = time.time()
        c.execute("UPDATE meta SET mtime = ?", (mtime,))

        # 7. Commit and wrap up (optimize the db).
        conn.commit()
        print("Optimizing DB...")
        nld.db.vacuum(conn=conn)
        games_added = nld.db.count_games(name, conn=conn)

    print(
        "Updated '%s' in %.2f sec. Size: %.2f MB, Games: %i"
        % (filename, mtime - stime, os.path.getsize(filename) / 1024**2, games_added)
    )


def add_nledata_directory(path, name, filename=nld.db.DB):
    """This function can be used to add any `nle_data` dataset to a database.

    Full games that are generated by an env such as:

        `gym.make("NetHackChallenge-v0", savedir="", save_ttyrec_every=k)`

    come with the following structure:

    The directory structure should look like:

        nle_data/
        ├── 20220414-112633_sdfd83ns/
        │   ├── nle.3968599.0.ttyrec.bz2
        │   ├── nle.3968599.k.ttyrec.bz2
        │   └── nle.3968599.xlogfile
        └── <date>-<time>_<random>/
            ├── nle.<process-id>.0.ttyrec.bz2
            ├── nle.<process-id>.k.ttyrec.bz2
            └── nle.<process-id>.xlogfile

    This algorithm should be deterministic and always return the same dataset
    from an empty database, regardless of environment.
    """
    with nld.db.db(filename=filename, rw=True) as conn:
        print("Adding dataset '%s' ('%s') to '%s' " % (name, path, filename))

        root = os.path.abspath(path)
        stime = time.time()
        c = conn.cursor()

        # 1. Check if the dataset name exists, and add the root.
        nld.db.create_dataset(name, root, conn=c, commit=False)

        # 2. For each xlogfile, read the games and take only the games that
        #   correspond to the ttyrecs that exist in the enclosing directory.
        for xlogfile in sorted(glob.iglob(path + "/*/*.xlogfile")):
            stem = xlogfile.replace(".xlogfile", ".*.ttyrec*.bz2")

            files = set(glob.iglob(stem))
            ttyrecnames = {f.split("/")[-1] for f in files}
            versions = {f.split("ttyrec")[-1].replace(".bz2", "") for f in files}
            assert len(versions) == 1, "Cannot add ttyrecs with different versions"
            version = versions.pop()

            if version == "":
                raise AssertionError(
                    "Ttyrec version (* in ttyrec*.bz2) must be > 1 for NLE data."
                )

            c.execute(
                "UPDATE roots SET ttyrec_version = ? WHERE dataset_name = ?",
                (int(version), name),
            )

            ttyrecs = []
            ttydir = str(os.path.dirname(xlogfile))

            _filter = partial(
                xlogfile_gen_filter,
                ttyrecs=ttyrecs,
                ttyrecnames=ttyrecnames,
                ttydir=ttydir,
            )

            # 3. Add games to `games` and `datasets` table.
            game_gen = game_data_generator(xlogfile, filter=_filter)
            insert_sql = f"""
                INSERT INTO games
                VALUES (NULL, {','.join('?' for _ in XLOGFILE_COLUMNS)} )
            """
            c.executemany(insert_sql, game_gen)
            gameids = nld.db.get_most_recent_games(c.rowcount, conn=c)
            nld.db.add_games(name, *gameids, conn=conn, commit=False)

            # 4. Add ttyrecs to `ttyrecs` table.
            # Note gameids are "most recently added" so must be reversed.
            ttyrec_gen = ttyrec_data_generator(ttyrecs, reversed(gameids), root)
            c.executemany("INSERT INTO ttyrecs VALUES (?,?,?,?,?)", ttyrec_gen)

        mtime = time.time()
        c.execute("UPDATE meta SET mtime = ?", (mtime,))

        conn.commit()
        nld.db.vacuum(conn=conn)
        games_added = nld.db.count_games(name, conn=conn)

    print(
        "Updated '%s' in %.2f sec. Size: %.2f MB, Games: %i"
        % (filename, mtime - stime, os.path.getsize(filename) / 1024**2, games_added)
    )


def ttyrec_data_generator(ttyrecs, gameids, root):
    last_gameid = None
    for path, gameid in zip(ttyrecs, gameids):
        if gameid != last_gameid:
            part = 0
        relpath = os.path.relpath(path, root)
        yield (
            relpath,
            part,
            os.path.getsize(path),
            os.path.getmtime(path),
            gameid,
        )
        part += 1
        last_gameid = gameid


def game_data_generator(xlogfile, filter=lambda x: x, separator="\t"):
    with open(xlogfile, "rb") as f:
        for line in filter(f.readlines()):
            game_data = collections.defaultdict(lambda: -1)
            for words in line.decode("latin-1").strip().split(separator):
                key, *var = words.split("=")
                game_data[key] = "=".join(var)

            if "while" in game_data:
                game_data["death"] += " while " + game_data["while"]

            yield tuple(ctype(game_data[key]) for key, ctype in XLOGFILE_COLUMNS)


def xlogfile_gen_filter(gen, ttyrecnames, ttyrecs, ttydir):
    """Filter lines of the xlogile, keeping files in `ttyrecnames` and storing the
    the accepted paths in `ttyrecs`."""
    # The `xlogfile` may have more rows than files in directory
    # due to 'save_ttyrec_every' option in env.py, so filter these out.
    # If we do find a file, we will save it to be added later.
    for line in gen:
        ttyrecname = line.decode("ascii").split("ttyrecname=")[-1].strip()
        if ttyrecname in ttyrecnames:
            ttyrecs.append(ttydir + "/" + ttyrecname)
            yield line



# Copyright (c) Facebook, Inc. and its affiliates.
import gymnasium as gym
from gymnasium.envs import registration

from nle.env.base import NLE, DUNGEON_SHAPE

_version = "v0"

for name in (
    "NetHack",
    "NetHackScore",
    "NetHackStaircase",
    "NetHackStaircasePet",
    "NetHackOracle",
    "NetHackGold",
    "NetHackEat",
    "NetHackScout",
    "NetHackChallenge",
):
    entry_point = "nle.env.tasks:" + name
    if name == "NetHack":
        entry_point = "nle.env.base:NLE"
    kwargs = {}
    if gym.__version__ >= "0.21":
        # Starting with version 0.21, gym wraps everything by the
        # OrderEnforcing wrapper by default (which isn't in gym.wrappers).
        # This breaks our seed() calls and some other code. Disable.
        kwargs["order_enforce"] = False
    registration.register(
        id="%s-%s" % (name, _version), entry_point=entry_point, **kwargs
    )


__all__ = ["NLE", "DUNGEON_SHAPE"]



# Copyright (c) Facebook, Inc. and its affiliates.
import enum
import logging
import os
import random
import sys
import tempfile
import time
import warnings
import weakref

import gymnasium as gym
import numpy as np

from nle import nethack

logger = logging.getLogger(__name__)

DUNGEON_SHAPE = nethack.DUNGEON_SHAPE


DEFAULT_MSG_PAD = 256
DEFAULT_INV_PAD = 55
DEFAULT_INVSTR_PAD = 80

ASCII_SPACE = ord(" ")
ASCII_y = ord("y")
ASCII_n = ord("n")
ASCII_ESC = nethack.C("[")

FULL_ACTIONS = nethack.USEFUL_ACTIONS

SKIP_EXCEPTIONS = (b"eat", b"attack", b"direction?", b"pray")

NLE_SPACE_ITEMS = (
    (
        "glyphs",
        gym.spaces.Box(
            low=0, high=nethack.MAX_GLYPH, **nethack.OBSERVATION_DESC["glyphs"]
        ),
    ),
    ("chars", gym.spaces.Box(low=0, high=255, **nethack.OBSERVATION_DESC["chars"])),
    ("colors", gym.spaces.Box(low=0, high=15, **nethack.OBSERVATION_DESC["colors"])),
    (
        "specials",
        gym.spaces.Box(low=0, high=255, **nethack.OBSERVATION_DESC["specials"]),
    ),
    (
        "blstats",
        gym.spaces.Box(
            low=np.iinfo(np.int32).min,
            high=np.iinfo(np.int32).max,
            **nethack.OBSERVATION_DESC["blstats"],
        ),
    ),
    (
        "message",
        gym.spaces.Box(
            low=np.iinfo(np.uint8).min,
            high=np.iinfo(np.uint8).max,
            **nethack.OBSERVATION_DESC["message"],
        ),
    ),
    (
        "program_state",
        gym.spaces.Box(
            low=np.iinfo(np.int32).min,
            high=np.iinfo(np.int32).max,
            **nethack.OBSERVATION_DESC["program_state"],
        ),
    ),
    (
        "internal",
        gym.spaces.Box(
            low=np.iinfo(np.int32).min,
            high=np.iinfo(np.int32).max,
            **nethack.OBSERVATION_DESC["internal"],
        ),
    ),
    (
        "inv_glyphs",
        gym.spaces.Box(
            low=0,
            high=nethack.MAX_GLYPH,
            **nethack.OBSERVATION_DESC["inv_glyphs"],
        ),
    ),
    (
        "inv_strs",
        gym.spaces.Box(low=0, high=255, **nethack.OBSERVATION_DESC["inv_strs"]),
    ),
    (
        "inv_letters",
        gym.spaces.Box(low=0, high=127, **nethack.OBSERVATION_DESC["inv_letters"]),
    ),
    (
        "inv_oclasses",
        gym.spaces.Box(
            low=0,
            high=nethack.MAXOCLASSES,
            **nethack.OBSERVATION_DESC["inv_oclasses"],
        ),
    ),
    (
        "screen_descriptions",
        gym.spaces.Box(
            low=0, high=127, **nethack.OBSERVATION_DESC["screen_descriptions"]
        ),
    ),
    (
        "tty_chars",
        gym.spaces.Box(low=0, high=255, **nethack.OBSERVATION_DESC["tty_chars"]),
    ),
    (
        "tty_colors",
        gym.spaces.Box(
            low=0,
            high=31,
            **nethack.OBSERVATION_DESC["tty_colors"],
        ),
    ),
    (
        "tty_cursor",
        gym.spaces.Box(low=0, high=255, **nethack.OBSERVATION_DESC["tty_cursor"]),
    ),
    (
        "misc",
        gym.spaces.Box(
            low=np.iinfo(np.int32).min,
            high=np.iinfo(np.int32).max,
            **nethack.OBSERVATION_DESC["misc"],
        ),
    ),
)


class NLE(gym.Env):
    """Standard NetHack Learning Environment.

    Implements a gym interface around `nethack.Nethack`.


    Examples:
        >>> env = NLE()
        >>> obs, reset_info = env.reset()
        >>> obs, reward, done, truncation, info = env.step(0)
        >>> env.render()
    """

    # Gymnasium expects an fps rate > 0 for render checks
    # but NetHack doesn't have any. Set it to 42, because
    # that is always the answer to life, the universe and
    # everything.
    metadata = {"render_modes": ["human", "ansi", "full"], "render_fps": 42}

    class StepStatus(enum.IntEnum):
        """Specifies the status of the terminal state.

        Note:
            User may redefine this class in subtasks to handle / categorize
            more terminal states.

            It is highly advised that, in such cases, the enums already defined
            in this object are replicated in some way. See `nle.env.tasks` for
            examples on how to do this right.
        """

        ABORTED = -1
        RUNNING = 0
        DEATH = 1

    def __init__(
        self,
        save_ttyrec_every=0,
        savedir=None,
        character="mon-hum-neu-mal",
        max_episode_steps=5000,
        observation_keys=(
            "glyphs",
            "chars",
            "colors",
            "specials",
            "blstats",
            "message",
            "inv_glyphs",
            "inv_strs",
            "inv_letters",
            "inv_oclasses",
            "screen_descriptions",
            "tty_chars",
            "tty_colors",
            "tty_cursor",
        ),
        actions=None,
        options=None,
        wizard=False,
        allow_all_yn_questions=False,
        allow_all_modes=False,
        spawn_monsters=True,
        render_mode="human",
    ):
        """Constructs a new NLE environment.

        Args:
            save_ttyrec_every: Integer, if 0, no ttyrecs (game recordings) will
                be saved. Otherwise, save a ttyrec every Nth episode.
            savedir (str or None): Path to save ttyrecs (game recordings) into,
                if save_ttyrec_every is nonzero. If nonempty string, interpreted
                as a path to a new or existing directory.
                If "" (empty string) or None, NLE choses a unique directory name.
            character (str): name of character. Defaults to "mon-hum-neu-mal".
            max_episode_steps (int): maximum amount of steps allowed before the
                game is forcefully quit. In such cases, ``info["end_status"]``
                will be equal to ``StepStatus.ABORTED``. Defaults to 5000.
            observation_keys (list): keys to use when creating the observation.
                Defaults to all.
            actions (list): list of actions. If None, the full action space will
                be used, i.e. ``nle.nethack.ACTIONS``. Defaults to None.
            options (list): list of game options to initialize Nethack. If None,
                Nethack will be initialized with the options found in
                ``nle.nethack.NETHACKOPTIONS`. Defaults to None.
            wizard (bool): activate wizard mode. Defaults to False.
            allow_all_yn_questions (bool):
                If set to True, no y/n questions in step() are declined.
                If set to False, only elements of SKIP_EXCEPTIONS are not declined.
                Defaults to False.
            allow_all_modes (bool):
                If set to True, do not decline menus, text input or auto 'MORE'.
                If set to False, only skip click through 'MORE' on death.
            spawn_monsters: If False, disables normal NetHack behavior to randomly
                create monsters.
            render_mode (str): mode used to render the screen. One of
                "human" | "ansi" | "full".
                Defaults to "human", i.e. what a human would see playing the game.
        """
        self.character = character
        self._max_episode_steps = max_episode_steps
        self._allow_all_yn_questions = allow_all_yn_questions
        self._allow_all_modes = allow_all_modes
        self._save_ttyrec_every = save_ttyrec_every
        self.render_mode = render_mode

        if actions is None:
            actions = FULL_ACTIONS
        self.actions = actions

        self.last_observation = ()

        try:
            if not save_ttyrec_every:
                self.savedir = None
            elif savedir:
                self.savedir = os.path.abspath(savedir)
                os.makedirs(self.savedir)
            else:  # Empty savedir: We create our unique savedir inside nle_data/.
                parent_dir = os.path.join(os.getcwd(), "nle_data")
                os.makedirs(parent_dir, exist_ok=True)
                self.savedir = tempfile.mkdtemp(
                    prefix=time.strftime("%Y%m%d-%H%M%S_"), dir=parent_dir
                )
        except FileExistsError:
            logger.info("Using existing savedir: %s", self.savedir)
        else:
            if self.savedir:
                logger.info("Created savedir: %s", self.savedir)
            else:
                logger.info("Not saving any NLE data.")

        self._observation_keys = list(observation_keys)

        if "internal" in self._observation_keys:
            logger.warn(
                "The 'internal' NLE observation was requested. "
                "This might contain data that shouldn't be available to agents."
            )

        # Observations we always need.
        for key in (
            "glyphs",
            "blstats",
            "message",
            "program_state",
            "internal",
        ):
            if key not in self._observation_keys:
                self._observation_keys.append(key)

        self._glyph_index = self._observation_keys.index("glyphs")
        self._blstats_index = self._observation_keys.index("blstats")
        self._message_index = self._observation_keys.index("message")
        self._program_state_index = self._observation_keys.index("program_state")
        self._internal_index = self._observation_keys.index("internal")

        self._original_observation_keys = observation_keys
        self._original_indices = tuple(
            self._observation_keys.index(key) for key in observation_keys
        )

        if self.savedir:
            ttyrec_version = ".ttyrec%i.bz2" % nethack.TTYREC_VERSION
            ttyrec_prefix = "nle.%i.%%i" % os.getpid()
            self._ttyrec_pattern = os.path.join(
                self.savedir, ttyrec_prefix + ttyrec_version
            )
            ttyrec = self._ttyrec_pattern % 0
            # Create an xlogfile with the same format of name.
            scoreprefix = ttyrec.replace("0" + ttyrec_version, "")
        else:
            ttyrec = None
            scoreprefix = None

        self.nethack = nethack.Nethack(
            observation_keys=self._observation_keys,
            options=options,
            playername="Agent-" + self.character,
            ttyrec=ttyrec,
            wizard=wizard,
            spawn_monsters=spawn_monsters,
            scoreprefix=scoreprefix,
        )
        self._close_nethack = weakref.finalize(self, self.nethack.close)

        self._random = random.SystemRandom()

        # -1 so that it's 0-based on first reset
        self._episode = -1

        space_dict = dict(NLE_SPACE_ITEMS)
        self.observation_space = gym.spaces.Dict(
            {key: space_dict[key] for key in observation_keys}
        )

        self.action_space = gym.spaces.Discrete(len(self.actions))

    def _get_observation(self, observation):
        return {
            key: observation[i]
            for key, i in zip(self._original_observation_keys, self._original_indices)
        }

    def _get_end_status(self, observation, done):
        if self._check_abort(observation):
            end_status = self.StepStatus.ABORTED
        else:
            end_status = self._is_episode_end(observation)
        return self.StepStatus(done or end_status)

    def _get_information(self, end_status):
        info = {}
        info["end_status"] = end_status
        info["is_ascended"] = self.nethack.how_done() == nethack.ASCENDED
        return info

    def print_action_meanings(self):
        for a_idx, a in enumerate(self.actions):
            print(a_idx, a)

    def _check_abort(self, observation):
        return self._steps >= self._max_episode_steps

    def step(self, action: int):
        """Steps the environment.

        Args:
            action (int): action integer as defined by ``self.action_space``.

        Returns:
            (dict, float, bool, dict): a tuple containing
                - (*dict*): an observation of the state; this will contain the keys
                  specified by ``self.observation_space``.
                - (*float*): a reward; see ``self._reward_fn`` to see how it is
                  specified.
                - (*bool*): True if the state is terminal, False otherwise.
                - (*bool*): True if the episode is truncated, False otherwise.
                - (*dict*): a dictionary of extra information (such as
                  `end_status`, i.e. a status info -- death, task win, etc. --
                  for the terminal state).
        """
        # Careful: By default we re-use Numpy arrays, so copy before!
        last_observation = tuple(a.copy() for a in self.last_observation)

        observation, done = self.nethack.step(self.actions[action])
        truncated = self._check_abort(observation)

        is_game_over = observation[self._program_state_index][0] == 1
        if is_game_over or not self._allow_all_modes:
            observation, done = self._perform_known_steps(
                observation, done, exceptions=True
            )

        self._steps += 1

        self.last_observation = observation

        end_status = self._get_end_status(observation, done)

        reward = float(
            self._reward_fn(last_observation, action, observation, end_status)
        )

        if end_status and not done:
            # Try to end the game nicely.
            self._quit_game(observation, done)
            done = True

        return (
            self._get_observation(observation),
            reward,
            done,
            truncated,
            self._get_information(end_status),
        )

    def _in_moveloop(self, observation):
        program_state = observation[self._program_state_index]
        return program_state[3]  # in_moveloop

    def reset(self, seed=None, options=None):
        """Resets the environment.

        Note:
            We attempt to manually navigate the first few menus so that the
            first seen state is ready to be acted upon by the user. This might
            fail in case Nethack is initialized with some uncommon options.

        Returns:
            (tuple) (Observation of the state as
                defined by `self.observation_space`,
                Extra game state information)
        """
        super().reset(seed=seed, options=options)
        self._episode += 1
        if self.savedir and self._episode % self._save_ttyrec_every == 0:
            new_ttyrec = self._ttyrec_pattern % self._episode
        else:
            new_ttyrec = None
        self.last_observation = self.nethack.reset(new_ttyrec, options=options)

        self._steps = 0
        done = False

        for _ in range(1000):
            # Get past initial phase of game. This should make sure
            # all the observations are present.
            if self._in_moveloop(self.last_observation):
                break
            # This fails if the agent picks up a scroll of scare
            # monster at the 0th turn and gets asked to name it.
            # Hence the defensive iteration above.
            # TODO: Detect this 'in_getlin' situation and handle it.
            self.last_observation, done = self.nethack.step(ASCII_SPACE)
            assert not done, "Game ended unexpectedly"
        else:
            warnings.warn(
                "Not in moveloop after 1000 tries, aborting (ttyrec: %s)." % new_ttyrec,
                stacklevel=2,
            )
            return self.reset(seed=seed, options=options)

        return self._get_observation(self.last_observation), self._get_information(
            self._get_end_status(self.last_observation, done)
        )

    def close(self):
        self._close_nethack()
        super().close()

    def seed(self, core=None, disp=None, reseed=False):
        """Sets the state of the NetHack RNGs after the next reset.

        NetHack 3.6 uses two RNGs, core and disp. This is to prevent
        RNG-manipulation by e.g. running into walls or other no-ops on the
        actual game state. This is a measure against "tool-assisted
        speedruns" (TAS). NLE can run in both NetHack's default mode and in
        TAS-friendly "no reseeding" if `reseed` is set to False, see below.

        Arguments:
            core [int or None]: Seed for the core RNG. If None, chose a random
                value.
            disp [int or None]: Seed for the disp (anti-TAS) RNG. If None, chose
                a random value.
            reseed [boolean]: As an Anti-TAS (automation) measure,
                NetHack 3.6 reseeds with true randomness every now and then. This
                flag enables or disables this behavior. If set to True, trajectories
                won't be reproducible.

        Returns:
            [tuple] The seeds supplied, in the form (core, disp, reseed).
        """
        if core is None:
            core = self._random.randrange(sys.maxsize)
        if disp is None:
            disp = self._random.randrange(sys.maxsize)
        self.nethack.set_initial_seeds(core, disp, reseed)
        return (core, disp, reseed)

    def get_seeds(self):
        """Returns current seeds.

        Returns:
            (tuple): Current NetHack (core, disp, reseed) state.
        """
        return self.nethack.get_current_seeds()

    def render(self):
        """Renders the state of the environment."""
        mode = self.render_mode

        if mode == "human":
            obs = self.last_observation
            tty_chars = obs[self._observation_keys.index("tty_chars")]
            tty_colors = obs[self._observation_keys.index("tty_colors")]
            tty_cursor = obs[self._observation_keys.index("tty_cursor")]
            print(nethack.tty_render(tty_chars, tty_colors, tty_cursor))
            return None

        if mode == "full":
            message_index = self._observation_keys.index("message")
            message = bytes(self.last_observation[message_index])
            print(message[: message.index(b"\0")])
            try:
                inv_strs_index = self._observation_keys.index("inv_strs")
                inv_letters_index = self._observation_keys.index("inv_letters")

                inv_strs = self.last_observation[inv_strs_index]
                inv_letters = self.last_observation[inv_letters_index]
                for letter, line in zip(inv_letters, inv_strs):
                    if np.all(line == 0):
                        break
                    print(
                        letter.tobytes().decode("utf-8"), line.tobytes().decode("utf-8")
                    )
            except ValueError:  # inv_strs/letters not used.
                pass

            chars = self.last_observation[self._observation_keys.index("chars")]
            colors = self.last_observation[self._observation_keys.index("colors")]
            print(nethack.tty_render(chars, colors))
            return None

        if mode in ("ansi", "string"):  # Misnomer: This is the least ANSI of them all.
            chars = self.last_observation[self._observation_keys.index("chars")]
            # TODO: Why return a string here but print in the other branches?
            return "\n".join([line.tobytes().decode("utf-8") for line in chars])

        return "\nInvalid render mode: " + mode

    def __repr__(self):
        return "<%s>" % self.__class__.__name__

    def _is_episode_end(self, observation):
        """Returns whether the episode has ended.

        Tasks may override this method to specify different conditions, so long
        as the return value has a well defined __int__ method (e.g. booleans,
        numerical types, enum.IntEnum) and that value is part of StepStatus.

        The return value will be stored into info["end_status"].
        """
        return self.StepStatus.RUNNING

    def _reward_fn(self, last_observation, action, observation, end_status):
        """Reward function. Difference between previous score and new score."""
        if not self.nethack.in_normal_game():
            # Before game started and after it ended blstats are zero.
            return 0.0
        old_score = last_observation[self._blstats_index][nethack.NLE_BL_SCORE]
        score = observation[self._blstats_index][nethack.NLE_BL_SCORE]
        del end_status  # Unused for "score" reward.
        del action  # Unused for "score reward.
        return score - old_score

    def _perform_known_steps(self, observation, done, exceptions=True):
        while not done:
            if observation[self._internal_index][3]:  # xwaitforspace
                observation, done = self.nethack.step(ASCII_SPACE)
                continue

            internal = observation[self._internal_index]
            in_yn_function = internal[1]
            in_getlin = internal[2]

            if in_getlin:  # Game asking for a line of text. We don't do that.
                observation, done = self.nethack.step(ASCII_ESC)
                continue

            if in_yn_function:  # Game asking for a single character.
                # Note: No auto-yes to final questions thanks to the disclose option.
                if exceptions:
                    # This causes an annoying unnecessary copy...
                    msg = bytes(observation[self._message_index])
                    # Do not skip some questions to allow agent to select
                    # stuff to eat, attack, and to select directions.

                    # do not skip if all allowed or the allowed message appears
                    if self._allow_all_yn_questions or any(
                        el in msg for el in SKIP_EXCEPTIONS
                    ):
                        break

                # Otherwise, auto-decline.
                observation, done = self.nethack.step(ASCII_ESC)

            break

        return observation, done

    def _quit_game(self, observation, done):
        """Smoothly quit a game."""
        # Get out of menus and windows.
        observation, done = self._perform_known_steps(
            observation, done, exceptions=False
        )

        if done:
            return

        # Quit the game.
        actions = [0x80 | ord("q"), ord("y")]  # M-q y
        for a in actions:
            observation, done = self.nethack.step(a)

        # Answer final questions.
        observation, done = self._perform_known_steps(
            observation, done, exceptions=False
        )

        if not done:
            # Somehow, the above logic failed us.
            warnings.warn(
                "Warning: smooth quitting of game failed, aborting.", stacklevel=2
            )



import bz2
import os
import re

import numpy as np
import pytest
from memory_profiler import memory_usage

from nle.dataset import Converter

# From
#   https://alt.org/nethack/trd/?file=https://s3.amazonaws.com/altorg/ttyrec/Anarchos/2020-10-03.17:27:10.ttyrec.bz2  # noqa: B950
TTYREC_2020 = "2020-10-03.17_27_10.ttyrec.bz2"
COLSROWS = "2020-10-03.17_27_10.rowscols.txt"
TIMESTAMPS = "2020-10-03.17_27_10.timestamps.txt.bz2"
FINALFRAME = "2020-10-03.17_27_10.finalframe.txt"
FINALFRAMECOLORS = "2020-10-03.17_27_10.finalframe.colors.txt"

# From
#  https://alt.org/nethack/trd/?file=https://s3.amazonaws.com/altorg/ttyrec/Qvazzler/2009-02-05.10:33:51.ttyrec.bz2  # noqa: B950
# This ttyrec uses IBMGraphics (code page 437)
TTYREC_IBMGRAPHICS = "2009-02-05.10_33_51.ttyrec.bz2"
TTYREC_IBMGRAPHICS_FRAME_10 = "2009-02-05.10_33_51.frame.10.txt"

# From
#  https://alt.org/nethack/trd/?file=https://s3.amazonaws.com/altorg/ttyrec/moo22/2018-09-27.00:20:39.ttyrec.bz2  # noqa: B950
TTYREC_2018 = "2018-09-27.00_20_39.ttyrec.bz2"

# From
#  https://alt.org/nethack/trd/?file=https://s3.amazonaws.com/altorg/ttyrec/mackeyth/2017-07-28.18:44:17.ttyrec.bz2  # noqa: B950
# This ttyrec uses DECGraphics (https://en.wikipedia.org/wiki/DEC_Special_Graphics)
TTYREC_DECGRAPHICS = "2017-07-28.18_44_17.ttyrec.bz2"
TTYREC_DECGRAPHICS_FRAME_10 = "2017-07-28.18_44_17.frame.10.txt"

# From
#  https://alt.org/nethack/trd/?file=https://s3.amazonaws.com/altorg/ttyrec/waIrus/2019-11-18.08:52:15.ttyrec.bz2  # noqa: B950
# This ttyrec uses unknown control sequences - it looks like maybe VT420 with rectangle control. `ESC [ %d ; %d ; %d ; %d z`  # noqa: B950
TTYREC_UNKGRAPHICS = "2019-11-18.08_52_15.ttyrec.bz2"
TTYREC_UNKGRAPHICS_FRAME_10 = "2019-11-18.08_52_15.frame.10.txt"

# From
#  https://alt.org/nethack/trd/?file=https://s3.amazonaws.com/altorg/ttyrec/fare/2012-02-16.21:09:51.ttyrec.bz2  # noqa: B950
# This ttyrec uses SHIFT IN and SHIFT OUT ASCII control sequences toggle DEC graphics set.  # noqa: B950
TTYREC_SHIFTIN = "2012-02-16.21_09_51.ttyrec.bz2"
TTYREC_SHIFTIN_FRAME_10 = "2012-02-16.21_09_51.frame.10.txt"

# Version 2 ttyrec
TTYREC_NLE_V2 = "nle.2734875.0.ttyrec2.bz2"
TTYREC_NLE_V2_FRAME_150 = "nle.2734875.0.frame.150.txt"
TTYREC_NLE_V2_ACTIONS = "nle.2734875.0.actions.txt"

# Version 3 ttyrec
TTYREC_NLE_V3 = "nle.2783801.0.ttyrec3.bz2"
TTYREC_NLE_V3_FRAME_44 = "nle.2783801.0.frame.44.txt"
TTYREC_NLE_V3_ACTIONS_SCORES = "nle.2783801.0.actions.scores.txt"

SEQ_LENGTH = 20
ROWS = 25
COLUMNS = 80

TTYREC_V1 = 1
TTYREC_V2 = 2
TTYREC_V3 = 3


def getfilename(filename):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), filename)


def load_and_convert(
    converter, ttyrec, chars, colors, cursors, timestamps, actions, scores
):
    converter.load_ttyrec(ttyrec)
    remaining = converter.convert(chars, colors, cursors, timestamps, actions, scores)
    while remaining == 0:
        remaining = converter.convert(
            chars, colors, cursors, timestamps, actions, scores
        )


class TestConverter:
    def test_is_loaded(self):
        converter = Converter(ROWS, COLUMNS, TTYREC_V1)
        assert not converter.is_loaded()
        converter.load_ttyrec(getfilename(TTYREC_2020))
        assert converter.is_loaded()

    def test_no_memleak(self):
        pytest.skip(
            "Test is flaky "
            "(https://github.com/heiner/nle/actions/runs/8943546654/job/24568502821)"
        )

        chars = np.zeros((SEQ_LENGTH, ROWS, COLUMNS), dtype=np.uint8)
        colors = np.zeros((SEQ_LENGTH, ROWS, COLUMNS), dtype=np.int8)
        cursors = np.zeros((SEQ_LENGTH, 2), dtype=np.int16)
        timestamps = np.zeros((SEQ_LENGTH,), dtype=np.int64)
        actions = np.zeros((SEQ_LENGTH), dtype=np.uint8)
        scores = np.zeros((SEQ_LENGTH), dtype=np.int32)
        ttyrec = getfilename(TTYREC_2020)
        converter = Converter(ROWS, COLUMNS, TTYREC_V1)

        def convert_n_times(n):
            for _ in range(n):
                load_and_convert(
                    converter,
                    ttyrec,
                    chars,
                    colors,
                    cursors,
                    timestamps,
                    actions,
                    scores,
                )

        memory_list = memory_usage((convert_n_times, (100,), {}))
        # After warmup the last few iterations should be constant memory
        memory_array = np.array(memory_list[5:])
        memory_change = (memory_array / memory_array[0]) - 1
        assert max(memory_change) < 0.002  # 0.2 per cent

    def test_ttyrec_with_extra_data(self, seq_length=500):
        converter = Converter(ROWS, COLUMNS, TTYREC_V1)

        chars = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.uint8)
        colors = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.int8)
        cursors = np.zeros((seq_length, 2), dtype=np.int16)
        timestamps = np.zeros((seq_length,), dtype=np.int64)
        actions = np.zeros((seq_length), dtype=np.uint8)
        scores = np.zeros((seq_length), dtype=np.int32)

        converter.load_ttyrec(getfilename(TTYREC_2018))
        remaining = converter.convert(
            chars, colors, cursors, timestamps, actions, scores
        )
        assert remaining == 165

    def test_data(self):
        converter = Converter(ROWS, COLUMNS, TTYREC_V1)
        assert converter.rows == ROWS
        assert converter.cols == COLUMNS

        chars = np.zeros((SEQ_LENGTH, ROWS, COLUMNS), dtype=np.uint8)
        colors = np.zeros((SEQ_LENGTH, ROWS, COLUMNS), dtype=np.int8)
        cursors = np.zeros((SEQ_LENGTH, 2), dtype=np.int16)
        timestamps = np.zeros((SEQ_LENGTH,), dtype=np.int64)
        actions = np.zeros((SEQ_LENGTH), dtype=np.uint8)
        scores = np.zeros((SEQ_LENGTH), dtype=np.int32)

        converter.load_ttyrec(getfilename(TTYREC_2020))
        with open(getfilename(COLSROWS)) as f:
            colsrows = [tuple(int(i) for i in line.split()) for line in f]

        with bz2.BZ2File(getfilename(TIMESTAMPS)) as f:
            saved_timestamps = [float(line) for line in f]

        num_frames = 0
        while True:
            remaining = converter.convert(
                chars, colors, cursors, timestamps, actions, scores
            )
            for (row, col), ts in zip(
                cursors[: SEQ_LENGTH - remaining], timestamps[: SEQ_LENGTH - remaining]
            ):
                assert (col, row) == colsrows[num_frames]
                assert pytest.approx(float(ts) / 1e6) == saved_timestamps[num_frames]
                # Cursor col == converter.cols when it is offscreen (ie cropped).
                assert 0 <= col < converter.cols + 1
                assert 0 <= row < converter.rows
                num_frames += 1
            if remaining > 0:
                break

        assert num_frames == len(colsrows)
        final_index = SEQ_LENGTH - remaining - 1
        with open(getfilename(FINALFRAME)) as f:
            for row, line in enumerate(f):
                actual = chars[final_index][row].tobytes().decode("utf-8").rstrip()
                assert actual == line.rstrip()
        with open(getfilename(FINALFRAMECOLORS)) as f:
            for row, line in enumerate(f):
                actual = ",".join(str(c) for c in colors[final_index][row])
                assert actual == line.rstrip()

    def test_noexist(self):
        fn = "/does/not/exist.txt"
        converter = Converter(25, 80, TTYREC_V1)
        with pytest.raises(
            FileNotFoundError, match=r"\[Errno 2\] No such file or directory: '%s'" % fn
        ):
            converter.load_ttyrec(fn)

    def test_illegal_buffers(self):
        converter = Converter(ROWS, COLUMNS, TTYREC_V1)
        converter.load_ttyrec(getfilename(TTYREC_2020))

        chars = np.zeros((10, ROWS, COLUMNS), dtype=np.uint8)
        colors = np.zeros((10, ROWS, COLUMNS), dtype=np.int8)
        cursors = np.zeros((11, 2), dtype=np.int16)
        actions = np.zeros((10), dtype=np.uint8)
        timestamps = np.zeros((10,), dtype=np.int64)
        scores = np.zeros((10), dtype=np.int32)

        with pytest.raises(
            ValueError,
            match=re.escape("Array has wrong shape (expected [ 10 2 ], got [ 11 2 ])"),
        ):
            converter.convert(chars, colors, cursors, timestamps, actions, scores)

        chars = np.zeros((10, ROWS, COLUMNS), dtype=np.uint8)
        colors = np.zeros((10, ROWS, COLUMNS - 1), dtype=np.int8)
        cursors = np.zeros((10, 2), dtype=np.int16)
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Array has wrong shape (expected [ 10 25 80 ], got [ 10 25 79 ])"
            ),
        ):
            converter.convert(chars, colors, cursors, timestamps, actions, scores)

        chars = np.zeros((10, ROWS - 1, COLUMNS), dtype=np.uint8)
        colors = np.zeros((10, ROWS - 1, COLUMNS), dtype=np.int8)
        cursors = np.zeros((10, 2), dtype=np.int16)
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Array has wrong shape (expected [ 10 25 80 ], got [ 10 24 80 ])"
            ),
        ):
            converter.convert(chars, colors, cursors, timestamps, actions, scores)

        chars = np.zeros((11, ROWS, COLUMNS), dtype=np.uint8)
        colors = np.zeros((11, ROWS, COLUMNS), dtype=np.int8)
        cursors = np.zeros((11, 2), dtype=np.int16)
        actions = np.zeros((10), dtype=np.uint8)
        timestamps = np.zeros((10,), dtype=np.int64)
        with pytest.raises(
            ValueError,
            match=re.escape("Array has wrong shape (expected [ 11 ], got [ 10 ])"),
        ):
            converter.convert(chars, colors, cursors, timestamps, actions, scores)

        chars = np.zeros((10, ROWS, COLUMNS), dtype=np.uint8)
        colors = np.zeros((10, ROWS, COLUMNS), dtype=np.int8)
        cursors = np.zeros((10, 3), dtype=np.int16)
        with pytest.raises(
            ValueError,
            match=re.escape("Array has wrong shape (expected [ 10 2 ], got [ 10 3 ])"),
        ):
            converter.convert(chars, colors, cursors, timestamps, actions, scores)

        chars = np.zeros((10, ROWS, COLUMNS, 7), dtype=np.uint8)
        cursors = np.zeros((10, 2), dtype=np.int16)
        with pytest.raises(
            ValueError,
            match=r"Array has wrong number of dimensions \(expected 3, got 4\)",
        ):
            converter.convert(chars, colors, cursors, timestamps, actions, scores)

        chars = np.zeros((10, ROWS, COLUMNS), dtype=np.uint8)
        cursors = np.zeros((10, 2, 1), dtype=np.int16)
        with pytest.raises(
            ValueError,
            match=r"Array has wrong number of dimensions \(expected 2, got 3\)",
        ):
            converter.convert(chars, colors, cursors, timestamps, actions, scores)

        chars = np.zeros((10, ROWS, COLUMNS), dtype=np.float32)
        cursors = np.zeros((10, 2), dtype=np.uint8)
        with pytest.raises(ValueError, match=r"Buffer dtype mismatch"):
            converter.convert(chars, colors, cursors, timestamps, actions, scores)

        chars = np.zeros((10, ROWS, COLUMNS), dtype=np.uint8)
        cursors = np.zeros((10, 2), dtype=np.uint8)
        with pytest.raises(ValueError, match=r"Buffer dtype mismatch"):
            converter.convert(chars, colors, cursors, timestamps, actions, scores)

        chars = "Hello"
        cursors = np.zeros((10, 2), dtype=np.int16)
        with pytest.raises(ValueError, match=r"Numpy array required"):
            converter.convert(chars, colors, cursors, timestamps, actions, scores)

        chars = np.uint(8)
        with pytest.raises(ValueError, match=r"Numpy array required"):
            converter.convert(chars, colors, cursors, timestamps, actions, scores)

        chars = np.zeros((10, ROWS, COLUMNS), dtype=np.uint8)
        timestamps = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        with pytest.raises(ValueError, match=r"Numpy array required"):
            converter.convert(chars, colors, cursors, timestamps, actions, scores)

    def test_ibm_graphics(self):
        seq_length = 10
        converter = Converter(ROWS, COLUMNS, TTYREC_V1)

        chars = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.uint8)
        colors = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.int8)
        cursors = np.zeros((seq_length, 2), dtype=np.int16)
        actions = np.zeros((seq_length), dtype=np.uint8)
        timestamps = np.zeros((seq_length,), dtype=np.int64)
        scores = np.zeros((seq_length), dtype=np.int32)

        converter.load_ttyrec(getfilename(TTYREC_IBMGRAPHICS))
        assert (
            converter.convert(chars, colors, cursors, timestamps, actions, scores) == 0
        )

        with open(getfilename(TTYREC_IBMGRAPHICS_FRAME_10)) as f:
            for row, line in enumerate(f):
                actual = chars[-1][row].tobytes().decode("utf-8").rstrip()
                assert actual == line.rstrip()

    def test_dec_graphics(self):
        seq_length = 10
        COLUMNS = 120
        converter = Converter(ROWS, COLUMNS, TTYREC_V1)

        chars = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.uint8)
        colors = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.int8)
        cursors = np.zeros((seq_length, 2), dtype=np.int16)
        actions = np.zeros((seq_length), dtype=np.uint8)
        timestamps = np.zeros((seq_length,), dtype=np.int64)
        scores = np.zeros((seq_length), dtype=np.int32)

        converter.load_ttyrec(getfilename(TTYREC_DECGRAPHICS))
        assert (
            converter.convert(chars, colors, cursors, timestamps, actions, scores) == 0
        )

        with open(getfilename(TTYREC_DECGRAPHICS_FRAME_10)) as f:
            for row, line in enumerate(f):
                actual = chars[-1][row].tobytes().decode("utf-8").rstrip()
                assert actual == line.rstrip()

    def test_unknown_control_sequence_graphics(self):
        seq_length = 10
        COLUMNS = 120
        converter = Converter(ROWS, COLUMNS, TTYREC_V1)

        chars = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.uint8)
        colors = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.int8)
        cursors = np.zeros((seq_length, 2), dtype=np.int16)
        actions = np.zeros((seq_length), dtype=np.uint8)
        timestamps = np.zeros((seq_length,), dtype=np.int64)
        scores = np.zeros((seq_length), dtype=np.int32)

        converter.load_ttyrec(getfilename(TTYREC_UNKGRAPHICS))
        assert (
            converter.convert(chars, colors, cursors, timestamps, actions, scores) == 0
        )

        with open(getfilename(TTYREC_UNKGRAPHICS_FRAME_10)) as f:
            for row, line in enumerate(f):
                actual = chars[-1][row].tobytes().decode("utf-8").rstrip()
                assert actual == line.rstrip()

    def test_shiftin_shiftout_graphics(self):
        seq_length = 10
        COLUMNS = 120
        converter = Converter(ROWS, COLUMNS, TTYREC_V1)

        chars = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.uint8)
        colors = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.int8)
        cursors = np.zeros((seq_length, 2), dtype=np.int16)
        actions = np.zeros((seq_length), dtype=np.uint8)
        timestamps = np.zeros((seq_length,), dtype=np.int64)
        scores = np.zeros((seq_length), dtype=np.int32)

        converter.load_ttyrec(getfilename(TTYREC_SHIFTIN))
        assert (
            converter.convert(chars, colors, cursors, timestamps, actions, scores) == 0
        )

        with open(getfilename(TTYREC_SHIFTIN_FRAME_10)) as f:
            for row, line in enumerate(f):
                actual = chars[9][row].tobytes().decode("utf-8").rstrip()
                assert actual == line.rstrip()

    def test_nle_v2_conversion(self):
        seq_length = 150
        COLUMNS = 120
        converter = Converter(ROWS, COLUMNS, TTYREC_V2)

        chars = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.uint8)
        colors = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.int8)
        cursors = np.zeros((seq_length, 2), dtype=np.int16)
        actions = np.zeros((seq_length), dtype=np.uint8)
        timestamps = np.zeros((seq_length,), dtype=np.int64)
        scores = np.zeros((seq_length), dtype=np.int32)

        converter.load_ttyrec(getfilename(TTYREC_NLE_V2))
        assert (
            converter.convert(chars, colors, cursors, timestamps, actions, scores) == 0
        )

        with open(getfilename(TTYREC_NLE_V2_FRAME_150)) as f:
            for row, line in enumerate(f):
                actual = chars[-1][row].tobytes().decode("utf-8").rstrip()
                assert actual == line.rstrip()

        with open(getfilename(TTYREC_NLE_V2_ACTIONS)) as f:
            assert " ".join("%i" % a for a in actions) == f.readlines()[0]

    def test_nle_v3_conversion(self):
        seq_length = 70
        COLUMNS = 120
        converter = Converter(ROWS, COLUMNS, TTYREC_V3)

        chars = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.uint8)
        colors = np.zeros((seq_length, ROWS, COLUMNS), dtype=np.int8)
        cursors = np.zeros((seq_length, 2), dtype=np.int16)
        actions = np.zeros((seq_length), dtype=np.uint8)
        timestamps = np.zeros((seq_length,), dtype=np.int64)
        scores = np.zeros((seq_length), dtype=np.int32)

        converter.load_ttyrec(getfilename(TTYREC_NLE_V3))
        assert (
            converter.convert(chars, colors, cursors, timestamps, actions, scores) == 1
        )

        with open(getfilename(TTYREC_NLE_V3_FRAME_44)) as f:
            for row, line in enumerate(f):
                actual = chars[43][row].tobytes().decode("utf-8").rstrip()
                assert actual == line.rstrip()

        with open(getfilename(TTYREC_NLE_V3_ACTIONS_SCORES)) as f:
            lines = f.readlines()
            assert " ".join("%i" % a for a in actions) == lines[0].rstrip()
            assert " ".join("%i" % s for s in scores) == lines[1].rstrip()



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
import argparse
import bz2
import gzip
import os
import re
import select
import struct
import termios
import time

parser = argparse.ArgumentParser()
parser.add_argument(
    "-1",
    "--no_input",
    action="store_true",
    help="Use ttyrec (not ttyrec2) format without input data",
)
parser.add_argument(
    "-n",
    "--no_wait",
    action="store_true",
    help="Ignore timestamp data and don't wait between frames",
)
parser.add_argument(
    "-f",
    "--fixed_frame_wait",
    action="store_true",
    help="Wait a fixed time between each frame.",
)
parser.add_argument(
    "-s", "--speed", default=1.0, type=float, help="Set playback speed multiplier"
)
parser.add_argument("-i", "--print_inputs", action="store_true", help="Print inputs")
parser.add_argument(
    "-p", "--peek", action="store_true", help="Peek mode (like tail -f)"
)
parser.add_argument(
    "filename", default="", type=str, nargs="?", help="tty record file, or - for stdin"
)
parser.add_argument("--start", default=0, type=int, help="Start at a specific frame")
parser.add_argument(
    "--end", default=float("inf"), type=int, help="Quit after a specific frame count"
)


def wait(diff, speed, drift=0.0):
    jump = 0

    if FLAGS.no_wait:
        return speed, drift, jump
    if FLAGS.fixed_frame_wait:
        diff = 0.2

    start = time.time()

    diff /= speed
    diff = max(diff - drift, 0.0)

    rlist, _, _ = select.select((0,), (), (), diff)
    if rlist:
        c = os.read(0, 1)
        if c in (b"+", b"f"):
            speed *= 2
        elif c in (b"-", b"s"):
            speed /= 2
        elif c == b"1":
            speed = 1
        elif c == b"h":
            jump = -1
        elif c == b"l":
            jump = 1
        elif c == b" ":
            select.select((0,), (), ())
        drift = 0.0
    else:
        stop = time.time()
        if diff == 0.0:
            diff = -drift
        drift = (stop - start) - diff
    return speed, drift, jump


def read_header(f, peek=False, no_input=False):
    while True:
        if no_input:
            header = f.read(12)
        else:
            header = f.read(13)

        if not header:
            if peek:
                # Never return, just wait for more data.
                time.sleep(0.25)
                continue
            return

        if no_input:
            sec, usec, length = struct.unpack("<iii", header)
            channel = 0
        else:
            sec, usec, length, channel = struct.unpack("<iiiB", header)

        if sec < 0 or usec < 0 or length < 1 or channel not in (0, 1):
            raise IOError("Illegal header %s" % ((sec, usec, length, channel),))
        timestamp = sec + usec * 1e-6
        yield timestamp, length, channel


CLRCODE = re.compile(rb"\033\[2?J")  # https://stackoverflow.com/a/37778152/1136208
INPUTS = ["Input %s (%i)" % (repr(chr(i)), i) for i in range(255)]


def process(f):
    speed = FLAGS.speed
    drift = 0.0
    prev = None
    jump = 0

    # Store the header positions, timestamps, and frame numbers before
    # clear screen commands.
    clrscreen = []

    lastpos = 0
    frame = [0, 0]  # Input and output frame.
    for timestamp, length, channel in read_header(
        f, peek=FLAGS.peek, no_input=FLAGS.no_input
    ):
        data = f.read(length)
        frame[channel] += 1

        # Wait for delta between frame timestamps, unless we are jumping
        # ahead or we are at the start. Also don't wait when we want to
        # skip some frames via --start. Note that we still render them.
        # We could also only start rendering from the last CLRSCREEN
        # before FLAGS.start, which would be potentially faster.
        if jump == 0 and prev is not None and frame[0] >= FLAGS.start:
            speed, drift, jump = wait(timestamp - prev, speed, drift)

        if channel == 0 and CLRCODE.search(data):
            clrscreen.append((lastpos, prev, [frame[0] - 1, frame[1]]))
            if jump > 0:
                jump = 0

        lastpos = f.seek(0, os.SEEK_CUR)
        prev = timestamp

        if jump > 0:
            continue
        if jump < 0:
            jump = 0

            if not clrscreen:
                clrscreen.append((0, None, [0, 0]))

            while clrscreen and timestamp - prev <= 0.005:
                # If we can jump, jump at least some amount.
                lastpos, prev, frame = clrscreen.pop()

            if lastpos:
                prev = timestamp  # Don't wait in next iteration

            f.seek(lastpos, os.SEEK_SET)
            continue

        if channel == 0:  # Output.
            os.write(1, data)
        elif channel == 1 and FLAGS.print_inputs:  # Input.
            os.write(
                1, b"\033[s\033[26;0f\033[37;1mFrame %d+%d:\033[0m " % tuple(frame)
            )  # Save Cursor & Jump to L26
            os.write(1, INPUTS[ord(data)].encode("ascii", "backslashreplace"))
            os.write(1, b" " * 32)
            os.write(1, b" \033[u")  # Jump back Cursor

        if frame[0] > FLAGS.end:
            return


def main():
    global FLAGS
    FLAGS = parser.parse_args()

    if not FLAGS.filename:
        parser.print_help()
        return

    if FLAGS.filename == "-":
        f = os.fdopen(os.dup(0), "rb")
        os.dup2(1, 0)
    elif os.path.splitext(FLAGS.filename)[1] in (".bz2", ".bzip2"):
        f = bz2.BZ2File(FLAGS.filename)
    elif os.path.splitext(FLAGS.filename)[1] in (".gz", ".gzip"):
        f = gzip.GzipFile(FLAGS.filename)
    else:
        f = open(FLAGS.filename, "rb")

    old = termios.tcgetattr(0)
    new = termios.tcgetattr(0)

    # Set to unbuffered, no echo.
    new[3] &= ~(termios.ICANON | termios.ECHO | termios.ECHONL)  # lflags
    termios.tcsetattr(0, termios.TCSANOW, new)
    try:
        if FLAGS.peek:
            # Skip all previous data.
            for _, length, _ in read_header(f, peek=False):
                f.seek(length, os.SEEK_CUR)
            FLAGS.no_wait = True
        process(f)
    except KeyboardInterrupt:
        pass
    finally:
        termios.tcsetattr(0, termios.TCSANOW, old)
        f.close()


if __name__ == "__main__":
    main()



import os
import sqlite3
import threading
from collections import defaultdict
from functools import partial

import numpy as np

from nle import _pyconverter as converter
from nle import dataset as nld


def convert_frames(
    converter,
    chars,
    colors,
    curs,
    timestamps,
    actions,
    scores,
    resets,
    gameids,
    load_fn,
):
    """Convert frames for a single batch entry.

    :param converter: A ttychars.Converter object with a loaded file.
    :param chars: Array of characters -  np.array(np.uint8) [ SEQ x ROW x COL]
    :param colors: Array of colors -  np.array(np.uint8) [ SEQ x ROW x COL]
    :param curs: Array of cursors -  np.array(np.int16) [ SEQ x 2 ]
    :param timestamps: Array of timestamps -  np.array(np.int64) [ SEQ ]
    :param actions: Array of actions at t in response to output at t
        - np.array(np.uint8) [ SEQ ]
    :param scores: Array of in-game scores -  np.array(np.int32) [ SEQ ]
    :param resets: Array of resets -  np.array(np.uint8) [ SEQ ]
    :param gameids: Array of the gameid of each frame - np.array(np.int32) [ SEQ ]
    :param load_fn: A callback that loads the next file into a converter:
        sig: load_fn(converter) -> bool is_success

    """

    resets[0] = 0
    while True:
        remaining = converter.convert(chars, colors, curs, timestamps, actions, scores)
        end = np.shape(chars)[0] - remaining

        resets[1:end] = 0
        gameids[:end] = converter.gameid
        if remaining == 0:
            return

        # There still space in the buffers; load a new ttyrec and carry on.
        chars = chars[-remaining:]
        colors = colors[-remaining:]
        curs = curs[-remaining:]
        timestamps = timestamps[-remaining:]
        actions = actions[-remaining:]
        scores = scores[-remaining:]
        resets = resets[-remaining:]
        gameids = gameids[-remaining:]
        if load_fn(converter):
            if converter.part == 0:
                resets[0] = 1
        else:
            chars.fill(0)
            colors.fill(0)
            curs.fill(0)
            timestamps.fill(0)
            actions.fill(0)
            scores.fill(0)
            resets.fill(0)
            gameids.fill(0)
            return


def _ttyrec_generator(
    batch_size, seq_length, rows, cols, load_fn, map_fn, ttyrec_version
):
    """A generator to fill minibatches with ttyrecs.

    :param load_fn: a function to load the next ttyrec into a converter.
       load_fn(ttyrecs.Converter conv) -> bool is_success
    :param map_fn: a function that maps a series of iterables through a fn.
       map_fn(fn, *iterables) -> <generator> (can use built-in map)

    """
    chars = np.zeros((batch_size, seq_length, rows, cols), dtype=np.uint8)
    colors = np.zeros((batch_size, seq_length, rows, cols), dtype=np.int8)
    cursors = np.zeros((batch_size, seq_length, 2), dtype=np.int16)
    timestamps = np.zeros((batch_size, seq_length), dtype=np.int64)
    actions = np.zeros((batch_size, seq_length), dtype=np.uint8)
    resets = np.zeros((batch_size, seq_length), dtype=np.uint8)
    gameids = np.zeros((batch_size, seq_length), dtype=np.int32)
    scores = np.zeros((batch_size, seq_length), dtype=np.int32)

    key_vals = [
        ("tty_chars", chars),
        ("tty_colors", colors),
        ("tty_cursor", cursors),
        ("timestamps", timestamps),
        ("done", resets),
        ("gameids", gameids),
    ]
    if ttyrec_version >= 2:
        key_vals.append(("keypresses", actions))
    if ttyrec_version >= 3:
        key_vals.append(("scores", scores))

    # Load initial gameids.
    converters = [
        converter.Converter(rows, cols, ttyrec_version) for _ in range(batch_size)
    ]
    assert all(load_fn(c) for c in converters), "Not enough ttyrecs to fill a batch!"

    # Convert (at least one minibatch)
    _convert_frames = partial(convert_frames, load_fn=load_fn)
    gameids[0, -1] = 1  # basically creating a "do-while" loop by setting an indicator
    while np.any(
        gameids[:, -1] != 0
    ):  # loop until only padding is found, i.e. end of data
        list(
            map_fn(
                _convert_frames,
                converters,
                chars,
                colors,
                cursors,
                timestamps,
                actions,
                scores,
                resets,
                gameids,
            )
        )

        yield dict(key_vals)


class TtyrecDataset:
    """Dataset object to allow iteration through the ttyrecs found in our ttyrec
    database.
    """

    def __init__(
        self,
        dataset_name,
        batch_size=128,
        seq_length=32,
        rows=24,
        cols=80,
        dbfilename=nld.db.DB,
        threadpool=None,
        gameids=None,
        shuffle=True,
        loop_forever=False,
        subselect_sql=None,
        subselect_sql_args=None,
    ):
        """
        An iterable dataset to load minibatches of NetHack games from compressed
        ttyrec*.bz2 files into numpy arrays. (shape: [batch_size, seq_length, ...])

        This class makes use of a sqlite3 database at `dbfilename` to find the
        metadata and the location of files in a dataset. It then uses these to
        create generators which convert the ttyrecs on the fly. Note that the
        dataset generators always reuse their numpy arrays, writing into the
        arrays instead of generating new ones. Methods to create and populate the
        db from an NLE directry can be found in `populate_db.py`.

        Example
        -------
            ```
            import nle.dataset as nld

            if not os.path.exists(nld.db.DB):
                nld.db.create()
                nld.populate_db.add_nledata_directory('path/to/nle_data', "data1")

            dataset = nld.TtyrecDataset("data1"):

            for mb in dataset:
                # NB: dataset reuses np arrays, for performance reasons
                print(mb)
            ```

        :param batch_size: Number of parallel games to load.
        :param seq_length: Number of frames to load per game.
        :param rows: Row size of the terminal screen.
        :param cols: Column size of the terminal screen.
        :param dbfilename: Path to the database file
        :param gameids: Use a subselection of games (gameids) only.
        :param shuffle: Shuffle the order of gameids before iterating through them.
        :param loop_forever: If true, cycle through gameids forever,
            insted of padding empty batch dims with 0's.
        :param subselect_sql: SQL Query to subselect games (gameids) using metadata
        :param subselect_sql_args: SQL Query Args to subselect games (gameids)
            using metadata.
        """
        self.batch_size = batch_size
        self.seq_length = seq_length
        self.rows = rows
        self.cols = cols

        self.shuffle = shuffle
        self.subselect_sql = subselect_sql
        self.loop_forever = loop_forever

        sql_args = (dataset_name,)
        core_sql = """
            SELECT ttyrecs.gameid, ttyrecs.part, ttyrecs.path
            FROM ttyrecs
            INNER JOIN datasets ON ttyrecs.gameid=datasets.gameid
            WHERE datasets.dataset_name=?"""

        meta_sql = """
            SELECT games.*
            FROM games
            INNER JOIN datasets ON games.gameid=datasets.gameid
            WHERE datasets.dataset_name=?"""

        if subselect_sql:
            path_select = """
                SELECT ttyrecs.gameid, ttyrecs.part, ttyrecs.path
                FROM ttyrecs
                INNER JOIN datasets ON ttyrecs.gameid=datasets.gameid
                WHERE datasets.dataset_name=?
                AND ttyrecs.gameid IN (%s)"""
            core_sql = path_select % subselect_sql

            meta_select = """
                SELECT games.*
                FROM games
                INNER JOIN datasets ON games.gameid=datasets.gameid
                WHERE datasets.dataset_name=?
                AND games.gameid IN (%s)"""
            meta_sql = meta_select % subselect_sql
            sql_args = subselect_sql_args if subselect_sql_args else ()
            sql_args = (dataset_name,) + sql_args

        self._games = defaultdict(list)
        self._meta = None  # Populate lazily.
        self.dbfilename = dbfilename
        with nld.db.connect(self.dbfilename) as conn:
            c = conn.cursor()

            for row in c.execute(core_sql, sql_args):
                self._games[row[0]].append(row[1:3])

            # Guarantee order is [part0, ..., partN] for multi-part games.
            for files in self._games.values():
                files.sort()

            self._rootpath = nld.db.get_root(dataset_name, conn)
            self._ttyrec_version = nld.db.get_ttyrec_version(dataset_name, conn)

        if gameids is None:
            gameids = self._games.keys()

        self._core_sql = core_sql
        self._meta_sql = meta_sql
        self._sql_args = sql_args
        self._gameids = list(gameids)
        self._threadpool = threadpool
        self._map = partial(self._threadpool.map, timeout=60) if threadpool else map

    def get_paths(self, gameid):
        return [path for _, path in self._games[gameid]]

    def get_meta(self, gameid):
        if self._meta is None:
            self.populate_metadata()
        if gameid not in self._meta:
            return None
        return self._meta[gameid][0]

    def get_meta_columns(self):
        if self._meta is None:
            self.populate_metadata()
        return self._meta_cols

    def populate_metadata(self):
        self._meta = defaultdict(list)
        with nld.db.connect(self.dbfilename) as conn:
            conn.row_factory = sqlite3.Row
            c = conn.cursor()
            for row in c.execute(self._meta_sql, self._sql_args):
                self._meta[row[0]].append(row)
            self._meta_cols = [desc[0] for desc in c.description]

    def _make_load_fn(self, gameids):
        """Make a closure to load the next gameid from the db into the converter."""
        lock = threading.Lock()
        count = [0]

        def _load_fn(converter):
            """Take the next part of the current game if available, else new game.
            Return True if load successful, else False."""
            gameid = converter.gameid
            part = converter.part + 1

            files = self.get_paths(gameid)
            if gameid == 0 or part >= len(files):
                with lock:
                    i = count[0]
                    count[0] += 1

                if (not self.loop_forever) and i >= len(gameids):
                    return False

                gameid = gameids[i % len(gameids)]
                files = self.get_paths(gameid)
                part = 0

            filename = files[part]
            filepath = os.path.join(self._rootpath, filename)
            converter.load_ttyrec(filepath, gameid=gameid, part=part)
            return True

        return _load_fn

    def __iter__(self):
        gameids = list(self._gameids)
        if self.shuffle:
            np.random.shuffle(gameids)

        return _ttyrec_generator(
            self.batch_size,
            self.seq_length,
            self.rows,
            self.cols,
            self._make_load_fn(gameids),
            self._map,
            self._ttyrec_version,
        )

    def get_ttyrecs(self, gameids, chunk_size=None):
        """Fetch data from a single episode, chunked into a sequence of tensors."""
        seq_length = chunk_size or self.seq_length
        mbs = []
        for mb in _ttyrec_generator(
            len(gameids),
            seq_length,
            self.rows,
            self.cols,
            self._make_load_fn(gameids),
            self._map,
            self._ttyrec_version,
        ):
            mbs.append({k: t.copy() for k, t in mb.items()})
        return mbs

    def get_ttyrec(self, gameid, chunk_size=None):
        return self.get_ttyrecs([gameid], chunk_size)



# Copyright (c) Facebook, Inc. and its affiliates.
import contextlib
import random
import sys
import termios
import timeit

import numpy as np

from nle import nethack

SELF_PLAY_EPISODES = 0


@contextlib.contextmanager
def no_echo(fd=0):
    old = termios.tcgetattr(fd)
    try:
        new = termios.tcgetattr(fd)
        new[3] &= ~termios.ICANON & ~termios.ECHO  # lflags
        termios.tcsetattr(fd, termios.TCSAFLUSH, new)
        yield
    finally:
        termios.tcsetattr(fd, termios.TCSAFLUSH, old)


def main():
    # MORE + compass directions + long compass directions.
    ACTIONS = [
        13,
        107,
        108,
        106,
        104,
        117,
        110,
        98,
        121,
        75,
        76,
        74,
        72,
        85,
        78,
        66,
        89,
    ]

    nle = nethack.Nethack(observation_keys=("chars", "blstats", "message", "inv_strs"))
    nle.reset()

    nle.step(ord("y"))
    nle.step(ord("y"))
    nle.step(ord("\n"))

    steps = 0
    total_start_time = timeit.default_timer()
    start_time = timeit.default_timer()
    start_steps = steps

    mean_sps = 0
    sps_n = 0

    num_episodes = 100
    max_episode_length = 1000

    for episode in range(num_episodes):
        for _ in range(max_episode_length):
            ch = random.choice(ACTIONS)
            _, done = nle.step(ch)
            if done:
                break

            steps += 1

            if steps % 10000 == 0:
                end_time = timeit.default_timer()
                sps = (steps - start_steps) / (end_time - start_time)
                sps_n += 1
                mean_sps += (sps - mean_sps) / sps_n
                print("%f SPS" % sps)
                start_time = end_time
                start_steps = steps
        if episode % 100 == 0:
            print("Finished episode %i after %i steps." % (episode + 1, steps))
        nle.reset()

    total_time = timeit.default_timer() - total_start_time
    print(
        "Finished after %i steps (%i episodes) and %.2f sec. Mean sps: %f"
        % (steps, num_episodes, total_time, steps / total_time)
    )

    for i in range(SELF_PLAY_EPISODES):
        print("Starting self-play episode", i)
        chars, blstats, message, inv_strs = nle.reset()
        done = False
        while not done:
            message = bytes(message)
            print(message)
            for line in inv_strs:
                if np.all(line == 0):
                    break
                print(line.tobytes().decode("utf-8"))
            for line in chars:
                print(line.tobytes().decode("utf-8"))
            print(blstats)
            try:
                with no_echo():
                    (chars, blstats, message, inv_strs), done = nle.step(
                        ord(sys.stdin.read(1))
                    )
            except KeyboardInterrupt:
                break


main()



# Copyright (c) Facebook, Inc. and its affiliates.
import argparse
import atexit
import datetime
import os
import re
import signal
import struct
import subprocess
import sys

parser = argparse.ArgumentParser()
parser.add_argument(
    "-1",
    "--no_input",
    action="store_true",
    help="Use ttyrec (not ttyrec2) format without input data",
)
parser.add_argument(
    "-c",
    "--no_unicode_csi",
    dest="unicode_csi",
    action="store_false",
    help="Don't display control sequence introducer with unicode symbol",
)
parser.add_argument("--start", default=0, type=int, help="Start at a specific frame")
parser.add_argument(
    "--end", default=float("inf"), type=int, help="Quit after a specific frame count"
)
parser.add_argument(
    "filename", default="", type=str, nargs="?", help="tty record file, or - for stdin"
)
parser.add_argument(
    "--no_pager",
    dest="use_pager",
    action="store_false",
    help="Don't pipe data to less",
)


def ttyframes(f, tty2=True):
    while True:
        if tty2:
            header = f.read(13)
        else:
            header = f.read(12)

        if not isinstance(header, bytes):
            raise IOError("File must be opened in binary mode.")

        if not header:
            return

        if tty2:
            sec, usec, length, channel = struct.unpack("<iiiB", header)
        else:
            sec, usec, length = struct.unpack("<iii", header)
            channel = 0

        if sec < 0 or usec < 0 or length < 0 or channel not in (0, 1, 2):
            raise IOError("Illegal header %s in %s" % ((sec, usec, length, channel), f))
        timestamp = sec + usec * 1e-6

        data = f.read(length)

        yield timestamp, channel, data


def getfile(filename):
    if filename == "-":
        f = os.fdopen(os.dup(0), "rb")
        os.dup2(1, 0)
        return f
    if os.path.splitext(filename)[1] in (".bz2", ".bzip2"):
        import bz2

        return bz2.BZ2File(filename)
    if os.path.splitext(filename)[1] in (".gz", ".gzip"):
        import gzip

        return gzip.GzipFile(filename)
    return open(filename, "rb")


def color(s, value):
    return "\033[%d;3%dm%s\033[0m" % (bool(value & 8), value & ~8, s)


CTRL_COLOR = 8  # Dark gray.
DEC_COLOR = 4  # Dark blue.
DEC_DATA_COLOR = 3  # Dark yellow.

FRAMECNT_COLOR = 2  # Dark green.
TIMESTAMP_COLOR = 7  # "Normal" color.
CHANNEL_COLOR = 2  # Dark green.
BRACES_COLOR = [11, 4, 5]  # Output: Bright yellow, input: dark blue, score: pink


# "Select Graphic Rendition" sequence.
COLOR_REGEX = re.compile(
    r"""
(?P<begin>
  (\\x1b\[)  # CSI (Control Sequence Introducer)
  (?P<bright>[0-9])
  ;3
  (?P<color>[0-9])
  m
)
(?P<data> .*?)
(?P<end> \\x1b\[0m)
""",
    re.X,
)

# Generic control sequence
CTRL_REGEX = re.compile(r"\\x1b\[ ([0-9];?) *.", re.X)

# DEC Special Graphics
DEC_REGEX = re.compile(
    r"""
(?P<begin> \\x1b\(0)
(?P<data> .*?)
(?P<end> \\x1b\(B)
""",
    re.X,
)

CSI_REGEX = re.compile(r"\\x1b\[")

CSI_UNICODE = "⌘"


def _colorsub(m):
    """Substitute color in `COLOR_REGEX`."""
    bright = int(m.group("bright"))
    value = int(m.group("color"))
    if bright:
        value |= 8
    return m.group("begin") + color(m.group("data"), value) + m.group("end")


def _ctrlsub(m):
    return color(m.group(0), CTRL_COLOR)


def _decsub(m):
    """Substitute for `DEC_REGEX`."""
    return (
        color(m.group("begin"), DEC_COLOR)
        + color(m.group("data"), DEC_DATA_COLOR)
        + color(m.group("end"), DEC_COLOR)
    )


def main():
    global FLAGS
    FLAGS = parser.parse_args()

    if not FLAGS.filename:
        parser.print_help()
        return

    if FLAGS.use_pager:
        setup_pager()

    frames = [0, 0, 0]
    with getfile(FLAGS.filename) as f:
        for timestamp, channel, data in ttyframes(f, tty2=not FLAGS.no_input):
            frames[channel] += 1

            if frames[0] < FLAGS.start:
                continue

            if frames[0] > FLAGS.end:
                return

            if channel == 0:
                arrow = "<-"
            elif channel == 1:
                char, *_ = struct.unpack("<B", data)
                data = chr(char).encode("ascii", "backslashreplace")
                arrow = "->"
            elif channel == 2:
                score, *_ = struct.unpack("<i", data)
                data = f"  {score} "
                arrow = "->"

            data = str(data)[2:-1]  # Strip b' and '

            data = re.sub(COLOR_REGEX, _colorsub, data)
            data = re.sub(CTRL_REGEX, _ctrlsub, data)
            data = re.sub(DEC_REGEX, _decsub, data)

            if FLAGS.unicode_csi:
                data = re.sub(CSI_REGEX, CSI_UNICODE, data)

            print(
                "%s %s %s%s%s%s"
                % (
                    color(str(frames), FRAMECNT_COLOR),
                    color(datetime.datetime.fromtimestamp(timestamp), TIMESTAMP_COLOR),
                    color(arrow, CHANNEL_COLOR),
                    color("{", BRACES_COLOR[channel]),
                    data,
                    color("}", BRACES_COLOR[channel]),
                )
            )


def setup_pager():
    """Pager logic. Compare pager.c in git."""
    if not os.isatty(1):
        return

    pager_env = os.environ.copy()
    pager_env["LESS"] = "FRX"
    pager = subprocess.Popen(["less"], bufsize=0, stdin=subprocess.PIPE, env=pager_env)
    os.dup2(pager.stdin.fileno(), 1)
    if os.isatty(2):
        os.dup2(pager.stdin.fileno(), 2)
    pager.stdin.close()

    def close_pager_fds():
        os.close(1)
        os.close(2)

    def wait_for_pager_atexit():
        sys.stdout.flush()
        sys.stderr.flush()
        close_pager_fds()
        pager.wait()

    signos = (
        signal.SIGINT,
        signal.SIGHUP,
        signal.SIGTERM,
        signal.SIGQUIT,
        signal.SIGPIPE,
    )

    handlers = {signo: signal.getsignal(signo) for signo in signos}

    def wait_for_pager_signal(signo, frame):
        close_pager_fds()
        pager.wait()
        for signo, handler in handlers.items():
            signal.signal(signo, handler)
        signal.raise_signal(signo)

    for signo in signos:
        signal.signal(signo, wait_for_pager_signal)
    atexit.register(wait_for_pager_atexit)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# List of environment variables used:
#
#  NLE_PACKAGE_NAME
#    Prefix of the generated package (defaults to "nle").
#
#  NLE_BUILD_RELEASE
#    If set, builds wheel (s)dist such as to prepare it for upload to PyPI.
#
#  HACKDIR
#    If set, install NetHack's data files in this directory.
#
#  USE_SEEDING
#    If set, seeding is disabled in all NLE environments.
#
import os
import pathlib
import subprocess
import sys
from distutils import spawn
from distutils import sysconfig

import setuptools
from setuptools.command import build_ext


class CMakeBuild(build_ext.build_ext):
    def run(self):  # Necessary for pip install -e.
        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        source_path = pathlib.Path(__file__).parent.resolve()
        output_path = (
            pathlib.Path(self.get_ext_fullpath(ext.name))
            .parent.joinpath("nle")
            .resolve()
        )
        hackdir_path = os.getenv("HACKDIR", output_path.joinpath("nethackdir"))

        os.makedirs(self.build_temp, exist_ok=True)
        build_type = "Debug" if self.debug else "Release"

        generator = "Ninja" if spawn.find_executable("ninja") else "Unix Makefiles"

        use_seeding = os.environ.get("USE_SEEDING", "ON")
        use_seeding = {"1": "ON", "0": "OFF"}.get(use_seeding, use_seeding.upper())

        cmake_cmd = [
            "cmake",
            str(source_path),
            "-G%s" % generator,
            "-DPYTHON_SRC_PARENT=%s" % source_path,
            # Tell cmake which Python we want.
            "-DPYTHON_EXECUTABLE=%s" % sys.executable,
            "-DCMAKE_BUILD_TYPE=%s" % build_type,
            "-DCMAKE_INSTALL_PREFIX=%s" % sys.base_prefix,
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=%s" % output_path,
            "-DHACKDIR=%s" % hackdir_path,
            "-DPYTHON_INCLUDE_DIR=%s" % sysconfig.get_python_inc(),
            "-DPYTHON_LIBRARY=%s" % sysconfig.get_config_var("LIBDIR"),
            "-DUSE_SEEDING=%s" % use_seeding,
        ]

        build_cmd = ["cmake", "--build", ".", "--parallel"]
        install_cmd = ["cmake", "--install", "."]

        try:
            subprocess.check_call(cmake_cmd, cwd=self.build_temp)
            subprocess.check_call(build_cmd, cwd=self.build_temp)
            # Installs nethackdir. TODO: Can't we do this with setuptools?
            subprocess.check_call(install_cmd, cwd=self.build_temp)
        except subprocess.CalledProcessError:
            # Don't obscure the error with a setuptools backtrace.
            sys.exit(1)


packages = [
    "nle",
    "nle.dataset",
    "nle.env",
    "nle.nethack",
    "nle.agent",
    "nle.scripts",
    "nle.tests",
]

entry_points = {
    "console_scripts": [
        "nle-play = nle.scripts.play:main",
        "nle-ttyrec = nle.scripts.ttyrec:main",
        "nle-ttyplay = nle.scripts.ttyplay:main",
        "nle-ttyplay2 = nle.scripts.ttyplay2:main",
        "nle-read-tty = nle.scripts.read_tty:main",
    ]
}


extras_deps = {
    "dev": [
        "pre-commit>=2.0.1",
        "isort>=5.13.2",
        "cmake_format>=0.6.10",
        "memory-profiler>=0.60.0",
        "pytest>=6.2.5",
        "pytest-benchmark>=3.4.1",
        "sphinx>=2.4.4",
        "sphinx-rtd-theme>=0.4.3",
        "setuptools>=69.5.1",
        "ruff>=0.4.3",
    ],
    "agent": ["torch>=1.3.1"],
}

extras_deps["all"] = [item for group in extras_deps.values() for item in group]


if __name__ == "__main__":
    package_name = os.getenv("NLE_PACKAGE_NAME", "nle")
    cwd = os.path.dirname(os.path.abspath(__file__))
    version = open("version.txt", "r").read().strip()
    sha = "Unknown"

    try:
        sha = (
            subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=cwd)
            .decode("ascii")
            .strip()
        )
    except subprocess.CalledProcessError:
        pass

    if sha != "Unknown" and not os.getenv("NLE_RELEASE_BUILD"):
        version += "+" + sha[:7]
    print("Building wheel {}-{}".format(package_name, version))

    version_path = os.path.join(cwd, "nle", "version.py")
    with open(version_path, "w") as f:
        f.write("__version__ = '{}'\n".format(version))
        f.write("git_version = {}\n".format(repr(sha)))

    with open("README.md") as f:
        long_description = f.read()

    setuptools.setup(
        name=package_name,
        version=version,
        description=(
            "The NetHack Learning Environment (NLE): "
            "a reinforcement learning environment based on NetHack"
        ),
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="The NLE Dev Team",
        url="https://github.com/heiner/nle",
        license="NetHack General Public License",
        entry_points=entry_points,
        packages=packages,
        ext_modules=[setuptools.Extension("nle", sources=[])],
        cmdclass={"build_ext": CMakeBuild},
        setup_requires=["pybind11>=2.2"],
        install_requires=["pybind11>=2.2", "numpy>=1.16", "gymnasium==0.29.1"],
        extras_require=extras_deps,
        python_requires=">=3.8",
        classifiers=[
            "License :: OSI Approved :: Nethack General Public License",
            "Development Status :: 5 - Production/Stable",
            "Operating System :: POSIX :: Linux",
            "Operating System :: MacOS",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: C",
            "Programming Language :: C++",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Games/Entertainment",
        ],
        zip_safe=False,
    )



# Copyright (c) Facebook, Inc. and its affiliates.

# TODO(heiner): Consider using __all__ as ruff suggests.
# ruff: noqa

# This is needed to register envs
import nle.env

from .version import __version__
